from ..registry import register_model  # noqa
from .model import Model, ModelConfig  # noqa
from .audio_classification import *
from .backbone import *
from .image2text import *
from .language_modeling import *
from .sequence_labeling import *
from .speech_recognition import *
from .text_classification import *
from .text_detection import *
from .text_embedding import *
from .text_generation import *



import os
from dataclasses import dataclass
from typing import Dict, List, Mapping, Tuple, Union

from ..builders import build_preprocessor
from ..configs import PreprocessorConfig
from ..constants import (
    DEFAULT_NORMALIZER_CONFIG_FILE,
    DEFAULT_PREPROCESSOR_SUBFOLDER,
    Backends,
)
from ..registry import register_preprocessor
from ..utils import Logger, is_backend_available
from .preprocessor import Preprocessor


if is_backend_available(Backends.TOKENIZERS):
    from tokenizers import Regex, normalizers

_required_backends = [
    Backends.TOKENIZERS,
]

logger = Logger(__name__)


@dataclass
class TextNormalizerConfig(PreprocessorConfig):
    name = "text_normalizer"
    replace_patterns: Union[List[Tuple[str, str]], List[List[str]], List[Dict[str, List]]] = None
    nfkd: bool = True
    nfkc: bool = True

    def __post_init__(self):
        # convert a list of dicts in replace_patterns to a list of tuples
        if self.replace_patterns is not None and len(self.replace_patterns):
            if isinstance(self.replace_patterns, Mapping):
                patterns = []
                for v in self.replace_patterns.values():
                    patterns += v
                self.replace_patterns = patterns


@register_preprocessor("text_normalizer", config_class=TextNormalizerConfig)
class TextNormalizer(Preprocessor):
    """
    A simple configurable text normalizer
    """

    required_backends = _required_backends

    preprocessor_subfolder = DEFAULT_PREPROCESSOR_SUBFOLDER
    normalizer_config_file = DEFAULT_NORMALIZER_CONFIG_FILE

    def __init__(self, config: TextNormalizerConfig, **kwargs):
        super().__init__(config, **kwargs)

    def __call__(
        self,
        inputs: Union[str, List[str]],
        replace_patterns: Union[List[Tuple[str, str]], List[List[str]]] = None,
        nfkd: bool = None,
        nfkc: bool = None,
        **kwargs,
    ):
        if isinstance(inputs, str):
            inputs = [inputs]

        nfkd = nfkd or self.config.nfkd
        nfkc = nfkc or self.config.nfkc
        replace_patterns = replace_patterns or self.config.replace_patterns

        if nfkd:
            inputs = [normalizers.NFKD().normalize_str(x) for x in inputs]
        if nfkc:
            inputs = [normalizers.NFKC().normalize_str(x) for x in inputs]

        if replace_patterns is not None:
            replacer = normalizers.Sequence(
                [normalizers.Replace(Regex(src), trg) for src, trg in self.config.replace_patterns]  # noqa
            )
            inputs = [replacer.normalize_str(x) for x in inputs]

        return inputs

    @classmethod
    def load(
        cls,
        hub_or_local_path,
        subfolder=None,
        config_filename=None,
        **kwargs
    ) -> "TextNormalizer":
        config_filename = config_filename or cls.normalizer_config_file
        subfolder = subfolder or cls.preprocessor_subfolder
        config = TextNormalizerConfig.load(
            hub_or_local_path,
            filename=config_filename,
            subfolder=subfolder,
        )
        normalizer = build_preprocessor(config.name, config, **kwargs)
        return normalizer

    def push_to_hub(
        self,
        repo_id,
        commit_message: str = None,
        subfolder: str = None,
        config_filename: str = None,
        private: bool = None,
    ):
        """
        Push normalizer config and other optional files to the Hub.

        Args:
            repo_id: Repo id on the Hub
            commit_message: Commit message
            subfolder: Optional subfolder for the normalizer
            config_filename: Optional normalizer config filename
            private: Whether to create a private repo if it does not exist already
        """
        subfolder = subfolder or self.preprocessor_subfolder
        config_filename = config_filename or self.normalizer_config_file

        if commit_message is None:
            commit_message = "Hezar: Upload normalizer"

        # upload config
        self.config.push_to_hub(
            repo_id=repo_id,
            filename=config_filename,
            subfolder=subfolder,
            commit_message=commit_message,
        )

    def save(
        self,
        path,
        subfolder=None,
        config_filename=None,
    ):
        config_filename = config_filename or self.normalizer_config_file
        subfolder = subfolder or self.preprocessor_subfolder
        os.makedirs(path, exist_ok=True)
        self.config.save(path, filename=config_filename, subfolder=subfolder)



import inspect
from re import sub
from time import perf_counter
from typing import Callable, Dict, List, Mapping, Union


__all__ = [
    "exec_timer",
    "snake_case",
    "permute_dict_list",
    "sanitize_params_for_fn",
]


class exec_timer:
    """
    A context manager that captures the execution time of all the operations inside it

    Examples:
        >>> with exec_timer() as timer:
        >>>     # operations here
        >>> print(timer.time)
    """

    def __enter__(self):
        self.time = perf_counter()
        return self

    def __exit__(self, type, value, traceback):
        self.time = perf_counter() - self.time


def snake_case(s):
    return "_".join(sub("([A-Z][a-z]+)", r" \1", sub("([A-Z]+)", r" \1", s.replace("-", " "))).split()).lower()


def permute_dict_list(dict_list: List[Dict]) -> Dict[str, List]:
    """
    Convert a list of dictionaries to a dictionary of lists

    Args:
        dict_list: Input list of dicts

    Returns:

    """
    if not len(dict_list):
        return {}
    d = {key: [x[key] for x in dict_list] for key in dict_list[0]}
    return d


def sanitize_params_for_fn(fn: Callable, params: Union[Dict, Mapping], **kwargs):
    """
    Given a dict of parameters or kwargs, you can figure out which ones must be passed to the `fn` based on its
    signature.

    Args:
         fn: The function object
         params: A dict of parameters with values
         kwargs: Keyword arguments that are merged with `params`

    Returns:
        The proper dict of parameters keys and values
    """
    params.update(**kwargs)
    params_signature = dict(inspect.signature(fn).parameters)
    fn_parameters = {p for p, v in params_signature.items() if v.kind not in (v.VAR_KEYWORD, v.VAR_POSITIONAL)}
    fn_params_names = set(fn_parameters)
    input_params = {p: params[p] for p in fn_params_names if p in params}
    return input_params



from dataclasses import dataclass

from ....configs import ModelConfig


@dataclass
class T5TextGenerationConfig(ModelConfig):
    name = "t5_text_generation"
    vocab_size: int = 32103
    d_model: int = 768
    d_kv: int = 64
    d_ff: int = 2048
    num_layers: int = 12
    num_decoder_layers: int = 12
    num_heads: int = 12
    relative_attention_num_buckets: int = 32
    relative_attention_max_distance: int = 128
    dropout_rate: float = 0.1
    layer_norm_epsilon: float = 1e-6
    initializer_factor: float = 1.0
    feed_forward_proj: str = "gated-gelu"
    is_encoder_decoder: bool = True
    tie_word_embeddings: bool = False
    use_cache: bool = True
    pad_token_id: int = 0
    decoder_start_token_id: int = 0
    eos_token_id: int = 1
    min_length: int = 0
    max_length: int = 100
    input_prefix: str = None



from dataclasses import dataclass, field
from enum import Enum
from typing import Dict

import torch
from datasets import load_dataset

from ...builders import build_preprocessor
from ...configs import DatasetConfig
from ...constants import Backends, TaskType
from ...preprocessors import ImageProcessorConfig, Tokenizer
from ...registry import register_dataset
from ...utils import Logger
from ..data_collators import CharLevelOCRDataCollator
from .dataset import Dataset


logger = Logger(__name__)

_required_backends = [Backends.SCIKIT]

fa_characters = [
    "", "آ", "ا", "ب", "پ", "ت", "ث", "ج", "چ", "ه", "خ", "د", "ذ", "ر", "ز", "ژ", "س", "ش",
    "ص", "ض", "ط", "ظ", "ع", "غ", "ف", "ق", "ک", "گ", "ل", "م", "ن", "و", "ه", "ی", " "
]
fa_numbers = ["۱", "۲", "۳", "۴", "۵", "۶", "۷", "۸", "۹", "۰"]
fa_special_characters = ["ء", "ؤ", "ئ", "أ", "ّ"]
fa_symbols = ["|", "(", ")", "+", "-", ":", "،", "!", ".", "؛", "=", "%", "؟"]
en_numbers = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "0"]
all_characters = fa_characters + fa_numbers + fa_special_characters + fa_symbols + en_numbers

ID2LABEL = dict(enumerate(all_characters))


class TextSplitType(str, Enum):
    CHAR_SPLIT = "char_split"  # mostly for char level ocr models
    TOKENIZE = "tokenize"  # mostly for transformer-based ocr models


@dataclass
class OCRDatasetConfig(DatasetConfig):
    """
    Configuration class for OCR datasets.

    Args:
        path (str): Path to the dataset.
        text_split_type (TextSplitType): Type of text splitting (CHAR_SPLIT or TOKENIZE).
        tokenizer_path (str): Path to the tokenizer file.
        id2label (Dict[int, str]): Mapping of label IDs to characters.
        text_column (str): Column name for text in the dataset.
        images_paths_column (str): Column name for image paths in the dataset.
        max_length (int): Maximum length of text.
        invalid_characters (list): List of invalid characters.
        reverse_text (bool): Whether to reverse the text.
        annotation_file (str): Path to the annotation file.
        test_split_size (float): Size of the test split.
        image_processor_config (ImageProcessorConfig): Configuration for image processing.

    """
    name = "ocr"
    task: TaskType = TaskType.IMAGE2TEXT
    path: str = None
    text_split_type: TextSplitType = TextSplitType.TOKENIZE
    tokenizer_path: str = None  # if left to None, text_split_type must be `char_split`
    id2label: Dict[int, str] = field(default_factory=lambda: ID2LABEL)
    text_column: str = "label"
    images_paths_column = "image_path"
    max_length: int = None
    invalid_characters: list = None
    reverse_text: bool = None
    annotation_file: str = None
    test_split_size: float = 0.2
    image_processor_config: ImageProcessorConfig = None


@register_dataset("ocr", config_class=OCRDatasetConfig)
class OCRDataset(Dataset):
    """
    General OCR dataset class.

    OCR dataset supports two types of image to text dataset. One is for tokenizer-based models in which the labels are
    tokens and the other is char-level models in which the labels are separated by character and the converted to ids.
    This behavior is specified by the `text_split_type` in config which can be either `tokenize` or `char_split`.

    """
    required_backends = _required_backends

    def __init__(self, config: OCRDatasetConfig, split=None, **kwargs):
        """
        Initializes a new OCRDataset instance.

        Args:
            config (OCRDatasetConfig): The configuration object for the dataset.
            split: Dataset split, defaults to None.
            **kwargs: Additional keyword arguments.

        """
        super().__init__(config=config, **kwargs)
        self.data = self._load(split)
        self.image_processor = build_preprocessor("image_processor", config=self.config.image_processor_config)
        if self.config.text_split_type == TextSplitType.TOKENIZE:
            if self.config.tokenizer_path is not None:
                self.tokenizer = Tokenizer.load(self.config.tokenizer_path)
                self.data_collator = None  # TODO resolve this in the future.
            else:
                raise ValueError("No `tokenizer_path` given although `text_split_type` is set to `tokenize`!")
        else:
            self.tokenizer = None
            self.data_collator = CharLevelOCRDataCollator()

    def __len__(self):
        """
        Returns the length of the dataset.

        Returns:
            int: The length of the dataset.

        """
        return len(self.data)

    def _is_label_valid(self, text):
        """
        Check if the label is valid based on the configured maximum length and invalid characters.

        Args:
            text (str): The text label.

        Returns:
            bool: True if the label is valid, False otherwise.

        """
        if len(text) > self.config.max_length:
            return False
        for char in text:
            if char not in self.config.id2label.values():
                return False
        return True

    def _load(self, split=None):
        """
        Load the dataset and clean up invalid samples.

        Args:
            split: Dataset split, defaults to None.

        Returns:
            Dataset: The cleaned dataset.

        """
        data = load_dataset(self.config.path)[split]
        # Cleanup dataset
        valid_indices = []
        invalid_indices = []
        for i, sample in enumerate(list(iter(data))):
            text, path = sample.values()
            if self._is_label_valid(text):
                valid_indices.append(i)
            else:
                invalid_indices.append(i)
        if len(invalid_indices):
            logger.warning(
                f"Invalid samples found in the dataset! Invalid indices: {invalid_indices}"
            )
        data = data.select(valid_indices)
        return data

    def _text_to_tensor(self, text):
        """
        Convert text to tensor based on the configured text_split_type.

        Args:
            text (str): The raw text.

        Returns:
            torch.Tensor: The output tensor.

        """
        # Tokenize text inputs if text_split_type is set to `tokenize`
        if self.config.text_split_type == TextSplitType.TOKENIZE:
            token_ids = self.tokenizer(text, padding="max_length", max_length=self.config.max_length)["token_ids"]
            # Make sure to ignore pad tokens by the loss function
            token_ids = [token_id if token_id != self.tokenizer.pad_token_id else -100 for token_id in token_ids]
            labels = torch.tensor(token_ids)
        # If split text is not tokenizer-based
        elif self.config.text_split_type == TextSplitType.CHAR_SPLIT:
            if self.config.reverse_text:
                text = text[::-1]
            label2id = {v: k for k, v in self.config.id2label.items()}
            labels = [label2id[x] for x in text]
            labels = torch.LongTensor(labels)
        else:
            raise ValueError(f"Invalid `text_split_type={self.config.text_split_type}`")

        return labels

    def __getitem__(self, index):
        """
        Get a specific item from the dataset.

        Args:
            index: Index of the item to retrieve.

        Returns:
            dict: The input data.

        """
        text, path = self.data[index].values()
        pixel_values = self.image_processor(path, return_tensors="pt")["pixel_values"][0]
        labels = self._text_to_tensor(text)
        inputs = {
            "pixel_values": pixel_values,
            "labels": labels,
        }
        return inputs



r"""
Builder functions are used to create an instance of a module e.g, models, preprocessors, etc. without having to import
their corresponding classes manually. These builders use modules' registries to do so. Every builder gets a name and
optional config or config kwargs to build the object.

Examples:

    >>> from hezar.builders import build_model
    >>> model = build_model('distilbert_text_classification', id2label={0: 'negative', 1: 'positive'})
    >>> print(model)

"""
from typing import Optional

from .configs import (
    DatasetConfig,
    EmbeddingConfig,
    MetricConfig,
    ModelConfig,
    PreprocessorConfig,
)
from .constants import SplitType
from .registry import (
    datasets_registry,
    embeddings_registry,
    metrics_registry,
    models_registry,
    preprocessors_registry,
)


__all__ = [
    "build_model",
    "build_dataset",
    "build_preprocessor",
    "build_embedding",
    "build_metric",
]


def build_model(name: str, config: Optional[ModelConfig] = None, **kwargs):
    """
    Build the model using its registry name. If config is None then the model is built using the default config. Notice
    that this function only builds the model and does not perform any weights loading/initialization unless these
    actions are done in the model's :func:`__init__` .

    Args:
        name (str): name of the model in the models' registry
        config (ModelConfig): a ModelConfig instance
        **kwargs: extra config parameters that are loaded to the model

    Returns:
        A Model instance
    """
    from .utils import list_available_models

    available_models = list_available_models()
    if name not in available_models:
        raise ValueError(f"Unknown model name: `{name}`!\nAvailable model names: {available_models}")
    config = config or models_registry[name].config_class()
    model = models_registry[name].module_class(config, **kwargs)
    return model


def build_preprocessor(name: str, config: Optional[PreprocessorConfig] = None, **kwargs):
    """
    Build the preprocessor using its registry name. If config is None then the preprocessor is built using the
    default config.

    Args:
        name (str): name of the preprocessor in the preprocessors' registry
        config (PreprocessorConfig): a PreprocessorConfig instance
        **kwargs: extra config parameters that are loaded to the preprocessor

    Returns:
        A Preprocessor instance
    """
    from .utils import list_available_preprocessors

    available_preprocessors = list_available_preprocessors()
    if name not in preprocessors_registry:
        raise ValueError(
            f"Unknown preprocessor name: `{name}`!\nAvailable preprocessor names: {available_preprocessors}"
        )
    config = config or preprocessors_registry[name].config_class()
    preprocessor = preprocessors_registry[name].module_class(config, **kwargs)
    return preprocessor


def build_dataset(name: str, config: Optional[DatasetConfig] = None, split: SplitType = None, **kwargs):
    """
    Build the dataset using its registry name. If config is None then the dataset is built using the
    default config.

    Args:
        name (str): name of the dataset in the datasets' registry
        config (DatasetConfig): a DatasetConfig instance
        split (str): Dataset split to load
        **kwargs: extra config parameters that are loaded to the dataset

    Returns:
        A Dataset instance
    """
    from .utils import list_available_datasets

    available_datasets = list_available_datasets()
    if name not in available_datasets:
        raise ValueError(
            f"Unknown dataset name: `{name}`!\nAvailable dataset names: {available_datasets}"
        )
    config = config or datasets_registry[name].config_class()
    dataset = datasets_registry[name].module_class(config, split, **kwargs)
    return dataset


def build_embedding(name: str, config: Optional[EmbeddingConfig] = None, **kwargs):
    """
    Build the embedding using its registry name. If config is None then the embedding is built using the
    default config.

    Args:
        name (str): Name of the embedding in the embeddings' registry
        config (EmbeddingConfig): An EmbeddingConfig instance
        **kwargs: Extra config parameters that are loaded to the embedding

    Returns:
        A Embedding instance
    """
    from .utils import list_available_embeddings

    available_embeddings = list_available_embeddings()
    if name not in available_embeddings:
        raise ValueError(
            f"Unknown embedding name: `{name}`!\nAvailable embedding names: {available_embeddings}"
        )
    config = config or embeddings_registry[name].config_class()
    embedding = embeddings_registry[name].module_class(config, **kwargs)
    return embedding


def build_metric(name: str, config: Optional[MetricConfig] = None, **kwargs):
    """
    Build the metric using its registry name. If config is None then the metric is built using the
    default config.

    Args:
        name (str): Name of the metric in the metrics' registry
        config (MetricConfig): A MetricConfig instance
        **kwargs: Extra config parameters that are loaded to the metric

    Returns:
        A Metric instance
    """
    from .utils import list_available_metrics

    available_metrics = list_available_metrics()
    if name not in available_metrics:
        raise ValueError(f"Unknown metric name: `{name}`!\nAvailable metric names: {available_metrics}")
    config = config or metrics_registry[name].config_class()
    metric = metrics_registry[name].module_class(config, **kwargs)
    return metric



from dataclasses import dataclass, field

from ....configs import ModelConfig
from ....constants import TaskType


@dataclass
class DistilBertSequenceLabelingConfig(ModelConfig):
    name = "distilbert_sequence_labeling"
    task: str = TaskType.SEQUENCE_LABELING
    num_labels: int = None
    id2label: dict = None
    activation: str = "gelu"
    attention_dropout: float = 0.1
    dim: int = 768
    dropout: float = 0.1
    initializer_range: float = 0.02
    max_position_embeddings: int = 512
    n_heads: int = 12
    n_layers: int = 6
    output_past: bool = True
    pad_token_id: int = 0
    qa_dropout: float = 0.1
    tie_weights_: bool = True
    vocab_size: int = 42000

    hidden_dropout_prob: float = 0.1
    use_cache: bool = True
    classifier_dropout: float = None
    prediction_skip_tokens: list = field(default_factory=lambda: ["[SEP]", "[CLS]"])



from dataclasses import dataclass
from typing import Iterable

from sklearn.metrics import accuracy_score

from ..configs import MetricConfig
from ..constants import Backends, MetricType
from ..registry import register_metric
from .metric import Metric


_required_backends = [
    Backends.SCIKIT,
]


@dataclass
class AccuracyConfig(MetricConfig):
    name = MetricType.ACCURACY
    normalize: bool = True
    sample_weight: Iterable[float] = None
    output_keys: tuple = ("accuracy",)


@register_metric("accuracy", config_class=AccuracyConfig)
class Accuracy(Metric):
    """
    Accuracy metric for numeric arrays backed by Scikit-learn's `accuracy_score`.

    Args:
        config (AccuracyConfig): Metric config object
        **kwargs: Extra config parameters passed as kwargs to update the `config`
    """
    required_backends = _required_backends

    def __init__(self, config: AccuracyConfig, **kwargs):
        super().__init__(config, **kwargs)

    def compute(
        self,
        predictions=None,
        targets=None,
        normalize=None,
        sample_weight=None,
        n_decimals=None,
        output_keys=None,
    ):
        """
        Compute the accuracy score for the given predictions against targets.

        Args:
            predictions: A list of prediction labels
            targets: A list of ground truth labels
            normalize: Whether to normalize the inputs or not
            sample_weight: Sample weight
            n_decimals: Floating point decimals for the final score
            output_keys: Filter the output keys

        Returns:
            A dictionary of the metric results
        """
        normalize = normalize or self.config.normalize
        sample_weight = sample_weight or self.config.sample_weight
        n_decimals = n_decimals or self.config.n_decimals
        output_keys = output_keys or self.config.output_keys

        score = accuracy_score(
            targets,
            predictions,
            normalize=normalize,
            sample_weight=sample_weight,
        )

        results = {"accuracy": round(float(score), n_decimals)}

        if output_keys:
            results = {k: v for k, v in results.items() if k in output_keys}

        return results



from dataclasses import dataclass
from typing import Dict

from ....configs import ModelConfig


@dataclass
class CRNNImage2TextConfig(ModelConfig):
    name = "crnn_image2text"
    id2label: Dict[int, str] = None
    blank_id: int = 0
    n_channels: int = 1
    image_height: int = 32
    image_width: int = 128
    map2seq_in_dim: int = 2048
    map2seq_out_dim: int = 64
    rnn_dim: int = 256
    reverse_prediction_text: bool = None



from typing import List, Union

import torch

from ....constants import Backends
from ....registry import register_model
from ....utils import is_backend_available
from ...model import Model
from ...model_outputs import TextGenerationOutput
from .gpt2_text_generation_config import GPT2TextGenerationConfig


if is_backend_available(Backends.TRANSFORMERS):
    from transformers import (
        GenerationConfig,
        GPT2Config,
        GPT2LMHeadModel,
    )

_required_backends = [Backends.TRANSFORMERS, Backends.TOKENIZERS]


@register_model("gpt2_text_generation", config_class=GPT2TextGenerationConfig)
class GPT2TextGeneration(Model):
    is_generative = True
    tokenizer_name = "bpe_tokenizer"
    required_backends = _required_backends
    loss_fn_name = "cross_entropy"

    def __init__(self, config: GPT2TextGenerationConfig, **kwargs):
        super().__init__(config, **kwargs)
        self.gpt2 = GPT2LMHeadModel(config=GPT2Config(**self.config))

    def forward(
        self,
        token_ids,
        past_key_values=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        **kwargs
    ):
        outputs = self.gpt2(
            input_ids=token_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            labels=None,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        return dict(outputs)

    def compute_loss(self, logits: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        labels = labels.to(logits.device)
        # Shift so that tokens < n predict n
        shift_logits = logits[..., :-1, :].contiguous()
        shift_labels = labels[..., 1:].contiguous()
        # Compute loss
        loss = self.criterion(shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))
        return loss

    def generate(self, token_ids, **kwargs):
        self.config.generation.update(kwargs or {})
        generation_config = GenerationConfig(**self.config.generation)
        generated_ids = self.gpt2.generate(token_ids, generation_config=generation_config)
        return generated_ids

    def preprocess(self, texts: Union[str, List[str]], **kwargs):
        tokenizer = self.preprocessor[self.tokenizer_name]
        inputs = tokenizer(texts, return_tensors="pt", device=self.device)
        return inputs

    def post_process(self, generated_ids: torch.Tensor):
        tokenizer = self.preprocessor[self.tokenizer_name]
        decoded_outputs = tokenizer.decode(generated_ids.cpu().numpy().tolist())
        outputs = [TextGenerationOutput(text=text) for text in decoded_outputs]
        return outputs



"""
RoBERTa base language model (HuggingFace Transformers) wrapped by a Hezar Model class
"""
from typing import List, Union

from ....constants import Backends
from ....models import Model
from ....registry import register_model
from ....utils import is_backend_available
from .roberta_config import RoBERTaConfig


if is_backend_available(Backends.TRANSFORMERS):
    from transformers import RobertaConfig, RobertaModel

_required_backends = [
    Backends.TRANSFORMERS,
    Backends.TOKENIZERS,
]


@register_model("roberta", config_class=RoBERTaConfig)
class RoBERTa(Model):
    required_backends = _required_backends
    tokenizer_name = "bpe_tokenizer"
    skip_keys_on_load = ["model.embeddings.position_ids", "roberta.embeddings.position_ids"]  # For older versions

    def __init__(self, config, **kwargs):
        super().__init__(config=config, **kwargs)
        self.roberta = RobertaModel(RobertaConfig(**self.config))

    def forward(
        self,
        token_ids,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        **kwargs,
    ):
        outputs = self.roberta(
            input_ids=token_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )
        return outputs

    def preprocess(self, inputs: Union[str, List[str]], **kwargs):
        if isinstance(inputs, str):
            inputs = [inputs]
        if "text_normalizer" in self.preprocessor:
            normalizer = self.preprocessor["text_normalizer"]
            inputs = normalizer(inputs)
        tokenizer = self.preprocessor[self.tokenizer_name]
        inputs = tokenizer(inputs, return_tensors="pt", device=self.device)
        return inputs

    def post_process(self, model_outputs, **kwargs):
        hidden_states = model_outputs.get("hidden_states", None)
        attentions = model_outputs.get("attentions", None)
        outputs = {
            "last_hidden_state": model_outputs.get("last_hidden_state"),
            "hidden_states": hidden_states,
            "attentions": attentions,
        }
        return outputs



import numpy as np
import torch


def _reconstruct(labels, blank=0):
    new_labels = []
    # merge same labels
    previous = None
    for label in labels:
        if label != previous:
            new_labels.append(label)
            previous = label
    # delete blank
    new_labels = [label for label in new_labels if label != blank]

    return new_labels


def greedy_decode(emission_log_prob, blank=0):
    labels = np.argmax(emission_log_prob, axis=-1)
    labels = _reconstruct(labels, blank=blank)
    return labels


def ctc_decode(log_probs, id2label=None, blank=0):
    emission_log_probs = np.transpose(log_probs.cpu().detach().numpy(), (1, 0, 2))
    batch_size, max_length, _ = emission_log_probs.shape

    # size of emission_log_probs: (batch, length, class)
    decoded_ids = []
    for emission_log_prob in emission_log_probs:
        ids = greedy_decode(emission_log_prob, blank=blank)
        if id2label:
            ids = [id2label[label] for label in ids]
        ids.extend([blank] * (max_length - len(ids)))
        decoded_ids.append(ids)
    return torch.tensor(decoded_ids)



from ..constants import RegistryType
from .common_utils import snake_case


__all__ = [
    "list_available_models",
    "list_available_preprocessors",
    "list_available_datasets",
    "list_available_metrics",
    "list_available_embeddings",
    "get_registry_point",
    "get_module_class",
    "get_module_config_class",
    "get_registry_key_by_module_class",
]


def list_available_models():
    registry = _get_registry_from_type(RegistryType.MODEL)

    return sorted(registry.keys())


def list_available_preprocessors():
    registry = _get_registry_from_type(RegistryType.PREPROCESSOR)

    return sorted(registry.keys())


def list_available_datasets():
    registry = _get_registry_from_type(RegistryType.DATASET)

    return sorted(registry.keys())


def list_available_metrics():
    registry = _get_registry_from_type(RegistryType.METRIC)

    return sorted(registry.keys())


def list_available_embeddings():
    registry = _get_registry_from_type(RegistryType.EMBEDDING)

    return sorted(registry.keys())


def _get_registry_from_type(registry_type: RegistryType):
    if registry_type == RegistryType.MODEL:
        from ..models import Model  # noqa
        from ..registry import models_registry  # noqa

        registry = models_registry

    elif registry_type == RegistryType.PREPROCESSOR:
        from ..preprocessors import Preprocessor  # noqa
        # Also import models since some preprocessors are in their own model module
        from ..models import Model # noqa
        from ..registry import preprocessors_registry  # noqa

        registry = preprocessors_registry

    elif registry_type == RegistryType.DATASET:
        from ..data import Dataset  # noqa
        from ..registry import datasets_registry  # noqa

        registry = datasets_registry

    elif registry_type == RegistryType.EMBEDDING:
        from ..embeddings import Embedding  # noqa
        from ..registry import embeddings_registry  # noqa

        registry = embeddings_registry

    elif registry_type == RegistryType.METRIC:
        from ..metrics import Metric  # noqa
        from ..registry import metrics_registry  # noqa

        registry = metrics_registry

    else:
        raise ValueError(f"Invalid `registry_type`: {registry_type}!")

    return registry


def get_registry_point(registry_key: str, registry_type: RegistryType):
    """
    Get the registry item by registry key name in a specific registry

    Args:
        registry_key: Module's name in the registry
        registry_type: Module's registry container type

    Returns:
        A Registry object
    """
    registry = _get_registry_from_type(registry_type)

    registry = registry[registry_key]
    return registry


def get_module_config_class(name: str, registry_type: RegistryType):
    """
    Get the config class for a given module based on its registry name.

    Args:
        name (str): Module's registry name
        registry_type (str): Registry type

    Returns:
        A class of type :class:`hezar.Config`
    """
    registry = _get_registry_from_type(registry_type)

    if name not in registry:
        return None

    config_cls = registry[name].config_class
    return config_cls


def get_module_class(name: str, registry_type: RegistryType):
    """
    Get module class based on registry name

    Args:
        name: Module's key name in its registry
        registry_type: Type of the module e.g, model, dataset, preprocessor, embedding, etc

    Returns:
        A class corresponding to the given module
    """
    registry = _get_registry_from_type(registry_type)

    name = snake_case(name)
    module_cls = registry[name].module_class
    return module_cls


def get_registry_key_by_module_class(module_class: type, registry_type: RegistryType):
    """
    Given the module class, return the registry key if exists

    Args:
        module_class: The module class (raw class, not the name or object)
        registry_type: The registry type

    Returns:
        The corresponding key for the class in its registry
    """
    registry = _get_registry_from_type(registry_type)
    key_values = {v.module_class.__name__: k for k, v in registry.items()}
    module_class_name = module_class.__name__
    if module_class_name not in key_values:
        raise KeyError(
            f"The requested {registry_type} class `{module_class_name}` does not exist "
            f"in the {registry_type}s registry!"
        )
    return key_values[module_class.__name__]



import os
from dataclasses import dataclass
from typing import List, Literal, Union

from ..constants import Backends
from ..registry import register_embedding
from ..utils import is_backend_available
from .embedding import Embedding, EmbeddingConfig

if is_backend_available(Backends.GENSIM):
    from gensim.models import word2vec

_required_backends = [
    Backends.GENSIM,
]


@dataclass
class Word2VecConfig(EmbeddingConfig):
    """
    Configuration class for Word2Vec embeddings.

    Attributes:
        name (str): Name of the embedding.
        dataset_path (str): Path to the dataset.
        vector_size (int): Size of the word vectors.
        window (int): Window size for context words.
        alpha (float): Learning rate.
        min_count (int): Ignores all words with a total frequency lower than this.
        seed (int): Seed for random number generation.
        workers (int): Number of workers for training.
        min_alpha (float): Minimum learning rate.
        cbow_mean (int): Constant for CBOW. Default is 1.
        epochs (int): Number of training epochs. Default is 5.
        train_algorithm (Literal["skipgram", "cbow"]): Training algorithm, either 'skipgram' or 'cbow'.
        save_format (Literal["binary", "text"]): Format for saving the model, either 'binary' or 'text'.
    """

    name = "word2vec"
    dataset_path: str = None
    vector_size: int = 300
    window: int = 5
    alpha: float = 0.025
    min_count: int = 1
    seed: int = 1
    workers: int = 3
    min_alpha: float = 0.0001
    cbow_mean = 1
    epochs = 5
    train_algorithm: Literal["skipgram", "cbow"] = "skipgram"
    save_format: Literal["binary", "text"] = "binary"


@register_embedding("word2vec", config_class=Word2VecConfig)
class Word2Vec(Embedding):
    """
    Word2Vec embedding class.

    Attributes:
        required_backends (List[Union[str, Backends]]): List of required backends.
    """

    required_backends = _required_backends

    def __init__(self, config: Word2VecConfig, embedding_file: str = None, vectors_file: str = None, **kwargs):
        """
        Initialize the Word2Vec embedding.

        Args:
            config (Word2VecConfig): Configuration object.
            embedding_file (str): Path to the embedding file.
            vectors_file (str): Path to the vectors file.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(config, embedding_file=embedding_file, vectors_file=vectors_file, **kwargs)

    def build(self):
        """
        Build the Word2Vec embedding model.

        Returns:
            gensim.models.Word2Vec: Word2Vec embedding model.
        """
        embedding_model = word2vec.Word2Vec(
            vector_size=self.config.vector_size,
            window=self.config.window,
            sg=1 if self.config.train_algorithm == "skipgram" else 0,
            workers=self.config.workers,
            alpha=self.config.alpha,
            min_alpha=self.config.min_alpha,
            min_count=self.config.min_count,
        )
        return embedding_model

    def from_file(self, embedding_path, vectors_path):
        """
        Load the Word2Vec embedding model from file.

        Args:
            embedding_path (str): Path to the embedding file.
            vectors_path (str): Path to the vectors file.

        Raises:
            ValueError: If vectors file is not found.

        Returns:
            gensim.models.Word2Vec: Loaded Word2Vec embedding model.
        """
        if not os.path.isfile(vectors_path):
            raise ValueError(
                f"Could not load or find vectors file at `{vectors_path}`! "
                f"Please make sure it's been downloaded properly!"
            )

        embedding_model = word2vec.Word2Vec.load(embedding_path)

        return embedding_model

    def train(
        self,
        dataset: List[str],
        epochs: int = 5,
    ):
        """
        Train the Word2Vec embedding model.

        Args:
            dataset (List[str]): List of sentences for training.
            epochs (int): Number of training epochs.
        """
        self.model.build_vocab(dataset)
        self.model.train(
            dataset,
            epochs=epochs,
            total_examples=self.model.corpus_count,
            total_words=self.model.corpus_total_words,
        )

    def save(
        self,
        path: Union[str, os.PathLike],
        filename: str = None,
        subfolder: str = None,
        save_config: bool = True,
        config_filename: str = None,
    ):
        """
        Save the Word2Vec embedding model to a specified path.

        Args:
            path (Union[str, os.PathLike]): Path to save the embedding model.
            filename (str): Name of the embedding file.
            subfolder (str): Subfolder within the path.
            save_config (bool): Whether to save the configuration.
            config_filename (str): Configuration file name.
        """
        filename = filename or self.filename
        config_filename = config_filename or self.config_filename
        subfolder = subfolder or self.subfolder

        save_dir = os.path.join(path, subfolder)
        os.makedirs(save_dir, exist_ok=True)
        self.config.save(path, config_filename, subfolder=subfolder)

        self.model.save(os.path.join(save_dir, filename))

    def similarity(self, word1: str, word2: str):
        """
        Get the similarity between two words.

        Args:
            word1 (str): First word.
            word2 (str): Second word.

        Returns:
            float: Similarity score.
        """
        if not isinstance(word1, str) or not isinstance(word2, str):
            raise ValueError(
                f"`Embedding.similarity()` takes two string objects!\n"
                f"`word1`: {type(word1)}, `word2`: {type(word2)}"
            )
        similarity = self.word_vectors.similarity(word1, word2)
        return similarity

    def doesnt_match(self, words: List[str]):
        """
        Get the word that doesn't match the others in a list.

        Args:
            words (List[str]): List of words.

        Returns:
            str: Word that doesn't match.
        """
        doesnt_match = self.word_vectors.doesnt_match(words)
        return doesnt_match

    def most_similar(self, word: str, top_n: int = 5):
        """
        Get the most similar words to a given word.

        Args:
            word (str): Input word.
            top_n (int): Number of similar words to retrieve.

        Returns:
            List[Dict[str, Union[str, float]]]: List of dictionaries containing 'word' and 'score'.
        """
        if not isinstance(word, str):
            raise ValueError(f"`word` must be `str`, got `{type(word)}`!")
        most_similar = self.word_vectors.most_similar(word, topn=top_n)
        most_similar = [{"word": word, "score": f"{score:.4f}"} for word, score in most_similar]
        return most_similar

    def get_normed_vectors(self):
        """
        Get normalized word vectors.

        Returns:
            Any: Normed word vectors.
        """
        normed_vectors = self.word_vectors.get_normed_vectors()
        return normed_vectors

    @property
    def word_vectors(self):
        """
        Get word vectors.

        Returns:
            gensim.models.keyedvectors.KeyedVectors: Word vectors.
        """
        return self.model.wv

    @property
    def vectors(self):
        """
        Get all vectors.

        Returns:
            numpy.ndarray: All vectors.
        """
        return self.model.wv.vectors

    @property
    def vocab(self):
        """
        Get vocabulary.

        Returns:
            Dict[str, int]: Vocabulary.
        """
        return self.model.wv.key_to_index



import os.path

import pandas as pd
from torch.utils.tensorboard import SummaryWriter


__all__ = [
    "AverageMeter",
    "MetricsTracker",
    "CSVLogger",
    "write_to_tensorboard",
]


class AverageMeter:
    """Compute and store the average and current value"""

    def __init__(self, name, fmt=":f"):
        self.name = name
        self.fmt = fmt
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = "{name} {val" + self.fmt + "} ({avg" + self.fmt + "})"
        return fmtstr.format(**self.__dict__)


class MetricsTracker:
    def __init__(self, metrics):
        self.metrics = metrics or []
        self.trackers = {}
        if len(self.metrics):
            for m in self.metrics.values():
                for metric_key in m.config.output_keys:
                    self.trackers[metric_key] = AverageMeter(metric_key)
        if "loss" not in self.trackers:
            self.trackers["loss"] = AverageMeter("loss")

    def update(self, results):
        for metric_name, tracker in self.trackers.items():
            tracker.update(results[metric_name])

    def reset(self):
        for tracker in self.trackers.values():
            tracker.reset()

    def avg(self):
        avg_results = {}
        for metric_name, tracker in self.trackers.items():
            avg_results[metric_name] = tracker.avg

        return avg_results


class CSVLogger:
    def __init__(self, logs_dir: str, csv_filename: str):
        self.save_path = os.path.join(logs_dir, csv_filename)
        self.df = pd.DataFrame({})

    def write(self, logs: dict, step: int):
        logs = {k: [v] for k, v in logs.items()}
        row = pd.DataFrame(logs)
        row.name = step
        self.df = pd.concat([self.df, row])
        self.df.to_csv(self.save_path, index=False)


def write_to_tensorboard(writer: SummaryWriter, logs: dict, step: int):
    for metric_name, value in logs.items():
        writer.add_scalar(metric_name, value, step)



from dataclasses import dataclass
from typing import Iterable

from sklearn.metrics import f1_score

from ..configs import MetricConfig
from ..constants import Backends, MetricType
from ..registry import register_metric
from .metric import Metric


_required_backends = [
    Backends.SCIKIT,
]


@dataclass
class F1Config(MetricConfig):
    """
    Configuration class for F1 metric.

    Args:
        name (MetricType): The type of metric, F1 in this case.
        pos_label (int): Label of the positive class.
        average (str): Type of averaging for the F1 score.
        sample_weight (Iterable[float]): Sample weights for the F1 score.
        output_keys (tuple): Keys to filter the metric results for output.
    """
    name = MetricType.F1
    pos_label: int = 1
    average: str = "macro"
    sample_weight: Iterable[float] = None
    output_keys: tuple = ("f1",)


@register_metric("f1", config_class=F1Config)
class F1(Metric):
    """
    F1 metric for evaluating classification performance using sklearn's `f1_score`.

    Args:
        config (F1Config): Metric configuration object.
        **kwargs: Extra configuration parameters passed as kwargs to update the `config`.
    """
    required_backends = _required_backends

    def __init__(self, config: F1Config, **kwargs):
        super().__init__(config, **kwargs)

    def compute(
        self,
        predictions=None,
        targets=None,
        labels=None,
        pos_label=1,
        average=None,
        sample_weight=None,
        zero_division="warn",
        n_decimals=None,
        output_keys=None,
    ):
        """
        Computes the F1 score for the given predictions against targets.

        Args:
            predictions: Predicted labels.
            targets: Ground truth labels.
            labels: List of labels to include in the calculation.
            pos_label (int): Label of the positive class.
            average (str): Type of averaging for the F1 score.
            sample_weight (Iterable[float]): Sample weights for the F1 score.
            zero_division (str): Strategy to use for zero-division, default is "warn".
            n_decimals (int): Number of decimals for the final score.
            output_keys (tuple): Filter the output keys.

        Returns:
            dict: A dictionary of the metric results, with keys specified by `output_keys`.
        """
        pos_label = pos_label or self.config.pos_label
        average = average or self.config.average
        sample_weight = sample_weight or self.config.sample_weight
        n_decimals = n_decimals or self.config.n_decimals
        output_keys = output_keys or self.config.output_keys

        score = f1_score(
            targets,
            predictions,
            labels=labels,
            pos_label=pos_label,
            average=average,
            sample_weight=sample_weight,
            zero_division=zero_division,
        )

        score = float(score) if score.size == 1 else score

        results = {"f1": round(float(score), n_decimals)}

        if output_keys:
            results = {k: v for k, v in results.items() if k in output_keys}

        return results



from ..registry import register_metric  # noqa
from .metric import Metric, MetricConfig  # noqa
from .accuracy import Accuracy, AccuracyConfig
from .bleu import BLEU, BLEUConfig
from .cer import CER, CERConfig
from .f1 import F1, F1Config
from .precision import Precision, PrecisionConfig
from .recall import Recall, RecallConfig
from .seqeval import Seqeval, SeqevalConfig
from .wer import WER, WERConfig
from .rouge import ROUGE, ROUGEConfig



from ..registry import register_embedding  # noqa
from .embedding import Embedding, EmbeddingConfig  # noqa
from .fasttext import FastText, FastTextConfig
from .word2vec import Word2Vec, Word2VecConfig



from typing import List, Union

import numpy as np
import torch

from ..builders import build_metric
from ..configs import MetricConfig
from ..constants import MetricType
from ..metrics import Metric
from .trainer_utils import MetricsTracker


__all__ = [
    "MetricsHandler",
    "TextClassificationMetricsHandler",
    "SequenceLabelingMetricsHandler",
    "SpeechRecognitionMetricsHandler",
    "AudioClassificationMetricsHandler",
    "Image2TextMetricHandler",
    "TextGenerationMetricsHandler",
]


class MetricsHandler:
    """
    Base metrics handler class for computing metrics. Subclasses must implement `compute_metrics` method based on
    their specific task.

    Args:
        metrics: A list of metrics (metric raw name or Metric object)
        model_config: Optional model config
        trainer_config: Optional trainer config
    """
    valid_metrics: List[MetricType] = []

    def __init__(self, metrics: List[Union[str, MetricType, Metric, MetricConfig]], trainer=None, **kwargs):
        self.metrics = self._setup_metrics(metrics)
        self.trainer = trainer
        self.tracker = MetricsTracker(self.metrics)

    def _setup_metrics(self, metrics):
        metrics_dict = {}
        metrics = metrics or []
        if not len(metrics):
            return metrics_dict
        for metric in metrics:
            if isinstance(metric, str):
                if metric not in self.valid_metrics:
                    raise ValueError(f"Invalid metric `{metric}`! Available metrics: {self.valid_metrics}")
                metrics_dict[metric] = build_metric(metric)
            elif isinstance(metric, MetricConfig):
                metrics_dict[metric.name] = build_metric(metric.name, config=metric)
            else:
                raise ValueError(f"Invalid metric type `{type(metric)}`! Available metrics: {self.valid_metrics}")
        return metrics_dict

    def compute_metrics(self, predictions, labels, **kwargs):
        """
        Given a batch of predictions and a batch of labels, compute all metrics

        Args:
            predictions: Predictions batch usually containing logits
            labels: Ground truth labels batch
        """
        raise NotImplementedError


class TextClassificationMetricsHandler(MetricsHandler):
    valid_metrics = [
        MetricType.ACCURACY,
        MetricType.RECALL,
        MetricType.PRECISION,
        MetricType.F1,
    ]

    def __init__(self, metrics: List[Union[str, MetricType, Metric, MetricConfig]], trainer=None):
        super().__init__(metrics=metrics, trainer=trainer)

    def compute_metrics(self, predictions, labels, **kwargs):
        predictions = np.array(predictions).argmax(1).flatten()
        labels = np.array(labels).flatten()
        results = {}
        for metric_name, metric in self.metrics.items():
            results.update(metric.compute(predictions, labels))
        return results


class SequenceLabelingMetricsHandler(MetricsHandler):
    valid_metrics = [MetricType.SEQEVAL]

    def __init__(self, metrics: List[Union[str, MetricType, Metric, MetricConfig]], trainer=None):
        super().__init__(metrics=metrics, trainer=trainer)

    def compute_metrics(self, predictions, labels, **kwargs):
        predictions = np.array(predictions).argmax(2).squeeze()
        labels = np.array(labels).squeeze()

        # Remove ignored index (special tokens) and append `B-` in the beginning for seqeval
        prefix = "" if self.trainer.train_dataset.config.is_iob_schema else "B-"
        true_predictions = [
            [f"{prefix}{self.trainer.model.config.id2label[p]}" for (p, l) in zip(prediction, label) if l != -100]
            for prediction, label in zip(predictions, labels)
        ]
        true_labels = [
            [f"{prefix}{self.trainer.model.config.id2label[l]}" for (p, l) in zip(prediction, label) if l != -100]
            for prediction, label in zip(predictions, labels)
        ]

        results = {}
        for metric_name, metric in self.metrics.items():
            x = metric.compute(true_predictions, true_labels)
            results.update(x)
        return results


class Image2TextMetricHandler(MetricsHandler):
    valid_metrics = [MetricType.CER]

    def __init__(self, metrics: List[Union[str, MetricType, Metric, MetricConfig]], trainer=None):
        super().__init__(metrics=metrics, trainer=trainer)

    def compute_metrics(self, predictions, labels, **kwargs):
        predictions = self.trainer.model.post_process(torch.tensor(predictions))
        labels = self.trainer.model.post_process(torch.tensor(labels))
        predictions = [x["text"] for x in predictions]
        labels = [x["text"] for x in labels]
        results = {}
        for metric_name, metric in self.metrics.items():
            x = metric.compute(predictions, labels)
            results.update(x)
        return results


class SpeechRecognitionMetricsHandler(MetricsHandler):
    def __init__(self, metrics: List[Union[str, MetricType, Metric, MetricConfig]], trainer=None):
        super().__init__(metrics=metrics, trainer=trainer)

    def compute_metrics(self, predictions, labels, **kwargs):
        return {}


class TextGenerationMetricsHandler(MetricsHandler):
    valid_metrics = [MetricType.ROUGE, MetricType.BLEU]

    def __init__(self, metrics: List[Union[str, MetricType, Metric, MetricConfig]], trainer=None):
        super().__init__(metrics=metrics, trainer=trainer)

    def compute_metrics(self, predictions, labels, **kwargs):
        predictions = self.trainer.model.post_process(torch.tensor(predictions))
        labels = self.trainer.model.post_process(torch.tensor(labels))
        predictions = [x["text"] for x in predictions]
        labels = [x["text"] for x in labels]
        results = {}
        for metric_name, metric in self.metrics.items():
            x = metric.compute(predictions, labels)
            results.update(x)
        return results


class AudioClassificationMetricsHandler(MetricsHandler):
    def __init__(self, metrics: List[Union[str, MetricType, Metric, MetricConfig]], trainer=None):
        super().__init__(metrics=metrics, trainer=trainer)

    def compute_metrics(self, predictions, labels, **kwargs):
        return {}



from dataclasses import dataclass
from typing import Mapping

import numpy as np

from ..builders import build_preprocessor
from ..configs import PreprocessorConfig
from ..constants import DEFAULT_FEATURE_EXTRACTOR_CONFIG_FILE
from ..utils import convert_batch_dict_dtype
from .preprocessor import Preprocessor


@dataclass
class AudioFeatureExtractorConfig(PreprocessorConfig):
    feature_size: int = None
    sampling_rate: int = 16000
    padding: str = None
    padding_value: float = 0.0
    padding_side: str = None


class AudioFeatureExtractor(Preprocessor):
    """
    Base class for all audio feature extractors.
    """

    model_input_name = "input_features"
    config_filename = DEFAULT_FEATURE_EXTRACTOR_CONFIG_FILE

    def __init__(self, config: AudioFeatureExtractorConfig, **kwargs):
        super().__init__(config=config, **kwargs)

    def __call__(self, inputs, **kwargs):
        raise NotImplementedError

    def pad(
        self,
        processed_features,
        padding=None,
        max_length=None,
        truncation=None,
        pad_to_multiple_of=None,
        return_attention_mask=None,
        return_tensors=None,
    ):
        """
        Pad input values / input vectors or a batch of input values / input vectors up to predefined length or to the
        max sequence length in the batch.

        Args:
            processed_features: Processed inputs to add padding to
            padding: Padding strategy which can be `longest`, `max_length`, `False`
            max_length: Max input length (Only effective if padding is `max_length` too, ignored otherwise)
            truncation: Whether to truncate long inputs or not
            pad_to_multiple_of: If set will pad the sequence to a multiple of the provided value.
            return_attention_mask:  Whether to return the attention mask.
            return_tensors: Tensors return type among `pt`, `np`, `list`
        """
        return_attention_mask = return_attention_mask or self.config.return_attention_mask
        padding = padding or self.config.padding
        if isinstance(processed_features, (list, tuple)) and isinstance(processed_features[0], Mapping):
            processed_features = {
                key: np.array([example[key] for example in processed_features]) for key in processed_features[0].keys()
            }
        if self.model_input_name not in processed_features:
            raise ValueError(
                f"Processed inputs must have a `{self.model_input_name}` key!\n"
                f"Provided keys: {list(processed_features.keys())}"
            )

        required_input = processed_features[self.model_input_name]

        if len(required_input) == 0:
            if return_attention_mask:
                processed_features["attention_mask"] = []
            return processed_features

        first_element = required_input[0]
        if isinstance(first_element, (list, tuple)):
            # first_element might be an empty list/tuple in some edge cases so we grab the first non-empty element.
            index = 0
            while len(required_input[index]) == 0:
                index += 1
            if index < len(required_input):
                first_element = required_input[index][0]

        # processed_features = convert_batch_dict_dtype(processed_features, dtype="np")
        padding_strategy = self._get_padding_strategy(padding=padding, max_length=max_length)

        required_input = processed_features[self.model_input_name]

        batch_size = len(required_input)
        if not all(len(v) == batch_size for v in processed_features.values()):
            raise ValueError("Some items in the output dictionary have a different batch size than others.")

        truncated_inputs = []
        for i in range(batch_size):
            inputs = {k: v[i] for k, v in processed_features.items()}
            # truncation
            inputs_slice = self._truncate(
                inputs,
                max_length=max_length,
                pad_to_multiple_of=pad_to_multiple_of,
                truncation=truncation,
            )
            truncated_inputs.append(inputs_slice)

        if padding_strategy == "longest":
            # make sure that `max_length` cannot be longer than the longest truncated length
            max_length = max(len(input_slice[self.model_input_name]) for input_slice in truncated_inputs)
            padding_strategy = "max_length"

        batch_outputs = {}
        for i in range(batch_size):
            # padding
            outputs = self._pad(
                truncated_inputs[i],
                max_length=max_length,
                padding_strategy=padding_strategy,
                pad_to_multiple_of=pad_to_multiple_of,
                return_attention_mask=return_attention_mask,
            )
            for key, value in outputs.items():
                if key not in batch_outputs:
                    batch_outputs[key] = []
                if value.dtype is np.dtype(np.float64):
                    value = value.astype(np.float32)
                batch_outputs[key].append(value)

        batch_outputs = {k: np.array(v) for k, v in batch_outputs.items()}

        padded_features = convert_batch_dict_dtype(batch_outputs, dtype=return_tensors)

        return padded_features

    def _pad(
        self,
        processed_features,
        max_length=None,
        padding_strategy=None,
        pad_to_multiple_of=None,
        return_attention_mask=None,
    ) -> dict:
        """
        Pad inputs based on padding strategy and max length

        Args:
            processed_features: Processed inputs to add padding to
            padding_strategy: Padding strategy which can be `longest`, `max_length`, `False`
            max_length: Max input length (Only effective if padding is `max_length` too, ignored otherwise)
            pad_to_multiple_of: If set will pad the sequence to a multiple of the provided value.
            return_attention_mask:  Whether to return the attention mask.

        Returns:
            Batch dict of the padded features
        """
        required_input = processed_features[self.model_input_name]

        if padding_strategy == "longest":
            max_length = len(required_input)

        if max_length is not None and pad_to_multiple_of is not None and (max_length % pad_to_multiple_of != 0):
            max_length = ((max_length // pad_to_multiple_of) + 1) * pad_to_multiple_of

        needs_to_be_padded = padding_strategy is not None and len(required_input) < max_length

        if return_attention_mask and "attention_mask" not in processed_features:
            processed_features["attention_mask"] = np.ones(len(required_input), dtype=np.int32)

        if needs_to_be_padded:
            difference = max_length - len(required_input)
            if self.config.padding_side == "right":
                if return_attention_mask:
                    processed_features["attention_mask"] = np.pad(processed_features["attention_mask"], (0, difference))
                padding_shape = ((0, difference), (0, 0)) if self.config.feature_size > 1 else (0, difference)
                processed_features[self.model_input_name] = np.pad(
                    required_input, padding_shape, "constant", constant_values=self.config.padding_value
                )
            elif self.config.padding_side == "left":
                if return_attention_mask:
                    processed_features["attention_mask"] = np.pad(processed_features["attention_mask"], (difference, 0))
                padding_shape = ((difference, 0), (0, 0)) if self.config.feature_size > 1 else (difference, 0)
                processed_features[self.model_input_name] = np.pad(
                    required_input, padding_shape, "constant", constant_values=self.config.padding_value
                )
            else:
                raise ValueError("Invalid padding strategy:" + str(self.config.padding_side))

        return processed_features

    def _get_padding_strategy(self, padding=False, max_length=None):
        """
        Find the correct padding strategy
        """
        if padding == "longest" or padding is True:
            padding_strategy = "longest"
            if self.config.padding_value is None:
                raise ValueError(f"Setting padding to `{padding_strategy}`, but `config.padding_value` is `None`!")

        elif padding == "max_length":
            if max_length is None:
                raise ValueError("Setting `padding=max_length` but leaving `max_length` as `None` is invalid!")
            padding_strategy = "max_length"

        else:
            padding_strategy = None

        return padding_strategy

    def _truncate(
        self,
        processed_features,
        max_length: int = None,
        pad_to_multiple_of: int = None,
        truncation: bool = None,
    ):
        """
        Truncate inputs to predefined length or max length in the batch

        Args:
            processed_features: Dictionary of input values
            max_length: maximum length of the returned list and optionally padding length
            pad_to_multiple_of: Integer if set will pad the sequence to a multiple of the provided value.
            truncation: Activates truncation to cut input sequences longer than `max_length` to `max_length`.
        """
        if not truncation:
            return processed_features
        elif truncation and max_length is None:
            raise ValueError("When setting ``truncation=True``, make sure that ``max_length`` is defined.")

        required_input = processed_features[self.model_input_name]

        # find `max_length` that fits `pad_to_multiple_of`
        if max_length is not None and pad_to_multiple_of is not None and (max_length % pad_to_multiple_of != 0):
            max_length = ((max_length // pad_to_multiple_of) + 1) * pad_to_multiple_of

        needs_to_be_truncated = len(required_input) > max_length

        if needs_to_be_truncated:
            processed_features[self.model_input_name] = processed_features[self.model_input_name][:max_length]
            if "attention_mask" in processed_features:
                processed_features["attention_mask"] = processed_features["attention_mask"][:max_length]

        return processed_features

    def save(
        self,
        path,
        subfolder=None,
        config_filename=None,
    ):
        """
        Save the feature extractor to the path. This normally is equal to only saving the
        `feature_extractor_config.yaml` file.

        Args:
            path: Main path to save the feature extractor files
            subfolder: Optional subfolder, defaults to `preprocessor`
            config_filename: Optional config file name, defaults to `feature_extractor_config.yaml`
        """
        subfolder = subfolder or self.preprocessor_subfolder
        config_filename = config_filename or self.config_filename

        self.config.save(path, filename=config_filename, subfolder=subfolder)

    def push_to_hub(
        self,
        repo_id,
        subfolder=None,
        commit_message=None,
        private=None,
        config_filename=None,
    ):
        """
        Push the feature extractor files to a repo on the Hub.

        Args:
            repo_id: Hub repo id
            subfolder: Subfolder to save, defaults to `self.preprocessor_subfolder` (`preprocessor`)
            commit_message: Commit message for the push
            private: If the repo does not exist already, specify whether the created repo must be private or not
            config_filename: Config filename, defaults to `self.config_filename` (`feature_extractor_config.yaml`)
        """
        subfolder = subfolder or self.preprocessor_subfolder
        config_filename = config_filename or self.config_filename

        if commit_message is None:
            commit_message = "Hezar: Upload feature extractor"

        self.config.push_to_hub(
            repo_id,
            subfolder=subfolder,
            filename=config_filename,
            private=private,
            commit_message=commit_message,
        )

    @classmethod
    def load(
        cls,
        hub_or_local_path,
        subfolder: str = None,
        config_filename: str = None,
        **kwargs,
    ):
        """
        Load a feature extractor from Hub or local path.

        Args:
            hub_or_local_path: Hub repo id or local path
            subfolder: Preprocessor subfolder path
            config_filename: Config file name
            **kwargs:

        Returns:
            A AudioFeatureExtractor object
        """
        subfolder = subfolder or cls.preprocessor_subfolder
        config_filename = config_filename or cls.config_filename

        config = AudioFeatureExtractorConfig.load(
            hub_or_local_path,
            subfolder=subfolder,
            filename=config_filename,
        )

        feature_extractor = build_preprocessor(config.name, config=config, **kwargs)

        return feature_extractor



"""
Configs are at the core of Hezar. All core modules like `Model`, `Preprocessor`, `Trainer`, etc. take their parameters
as a config container which is an instance of `Config` or its derivatives. A `Config` is a Python dataclass with
auxiliary methods for loading, saving, uploading to the hub and etc.

Examples:
    >>> from hezar.configs import ModelConfig
    >>> config = ModelConfig.load("hezarai/bert-base-fa")

    >>> from hezar.models import BertLMConfig
    >>> bert_config = BertLMConfig(vocab_size=50000, hidden_size=768)
    >>> bert_config.save("saved/bert", filename="model_config.yaml")
    >>> bert_config.push_to_hub("hezarai/bert-custom", filename="model_config.yaml")
"""
import os
import tempfile
from dataclasses import asdict, dataclass, field
from enum import Enum
from pprint import pformat
from typing import Dict, List, Optional, Tuple, Union

from huggingface_hub import create_repo, hf_hub_download, upload_file
from omegaconf import DictConfig, OmegaConf

from .constants import DEFAULT_MODEL_CONFIG_FILE, HEZAR_CACHE_DIR, ConfigType, TaskType
from .utils import Logger, get_module_config_class


__all__ = [
    "Config",
    "ModelConfig",
    "PreprocessorConfig",
    "TrainerConfig",
    "DatasetConfig",
    "EmbeddingConfig",
    "MetricConfig",
]

logger = Logger(__name__)

CONFIG_CLASS_VARS = ["name", "config_type"]

_config_to_type_mapping = {
    "ModelConfig": ConfigType.MODEL,
    "PreprocessorConfig": ConfigType.PREPROCESSOR,
    "TrainerConfig": ConfigType.TRAINER,
    "DatasetConfig": ConfigType.DATASET,
    "EmbeddingConfig": ConfigType.EMBEDDING,
    "CriterionConfig": ConfigType.CRITERION,
    "MetricConfig": ConfigType.METRIC,
}
_type_to_config_mapping = {v: k for k, v in _config_to_type_mapping.items()}


@dataclass
class Config:
    """
    Base class for all configs in Hezar.

    All configs are simple dataclasses that have some customized functionalities to manage their attributes. There are
    also some Hezar specific methods: `load`, `save` and `push_to_hub`.

    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.BASE)

    def __post_init__(self):
        if "name" in self.__annotations__:
            if self.__dataclass_fields__["name"].init == True:  # noqa
                raise ValueError(
                    f"The parameter `name` in a config should be either non-initable or unannotated! "
                    f"\nYou should define it as either:\n"
                    f"`name = {self.name}`"
                    f" or "
                    f"`name: str = field(default={self.name}, init=False)`"
                )
        # If `name` is defined raw, redefine it so that it's registered as a dataclass field
        elif "name" not in self.__annotations__ and self.name is not None:
            self.name: str = self.name

        # Convert enums to values
        for param, value in self.dict().items():
            if isinstance(getattr(self, param), Enum):
                setattr(self, param, getattr(self, param).value)

    def __str__(self):
        return pformat(self.dict())

    def __getitem__(self, item):
        try:
            return self.dict()[item]
        except KeyError:
            raise AttributeError(f"`{self.__class__.__name__}` does not have the parameter `{item}`!")

    def __len__(self):
        return len(self.dict())

    def __iter__(self):
        return iter(self.dict())

    def dict(self):
        """
        Returns the config object as a dictionary (works on nested dataclasses too)

        Returns:
            The config object as a dictionary
        """
        return asdict(self)

    def keys(self):
        return list(self.dict().keys())

    def get(self, key, default=None):
        return getattr(self, key, default)

    def update(self, d: dict, **kwargs):
        """
        Update config with a given dictionary or keyword arguments. If a key does not exist in the attributes, prints a
        warning but sets it anyway.

        Args:
            d: A dictionary
            **kwargs: Key/value pairs in the form of keyword arguments

        Returns:
            The config object itself but the operation happens in-place anyway
        """
        d.update(kwargs)
        for k, v in d.items():
            if k not in self.__annotations__.keys():
                logger.warning(f"`{str(self.__class__.__name__)}` does not take `{k}` as a config parameter!")
            setattr(self, k, v)
        return self

    @classmethod
    def load(
        cls,
        hub_or_local_path: Union[str, os.PathLike],
        filename: Optional[str] = None,
        subfolder: Optional[str] = None,
        repo_type=None,
        **kwargs,
    ) -> "Config":
        """
        Load config from Hub or locally if it already exists on disk (handled by HfApi)

        Args:
            hub_or_local_path: Local or Hub path for the config
            filename: Configuration filename
            subfolder: Optional subfolder path where the config is in
            repo_type: Repo type e.g, model, dataset, etc
            **kwargs: Manual config parameters to override

        Returns:
            A Config instance
        """
        filename = filename or DEFAULT_MODEL_CONFIG_FILE
        subfolder = subfolder or ""

        config_path = os.path.join(hub_or_local_path, subfolder, filename)
        is_local = os.path.isfile(config_path)
        if os.path.isdir(hub_or_local_path) and not is_local:
            raise EnvironmentError(
                f"Path `{hub_or_local_path}` exists locally but the config file {filename} is missing!"
            )
        # if the file or repo_id does not exist locally, load from the Hub
        if not is_local:
            config_path = hf_hub_download(
                hub_or_local_path,
                filename=filename,
                subfolder=subfolder,
                cache_dir=HEZAR_CACHE_DIR,
                repo_type=repo_type,
            )
        # Load config file and convert to dictionary
        dict_config = OmegaConf.load(config_path)
        config = OmegaConf.to_container(dict_config)
        # Check if config_type in the file and class are equal
        config_type = config.get("config_type", None)
        if config_type is None:
            logger.warning(f"`config_type` parameter in `{filename}` is `None` or does not exist!")
        elif config_type in _config_to_type_mapping.values():
            if config_type != cls.config_type:
                raise ValueError(
                    f"The `config_type` for `{cls.__name__}` is `{cls.config_type}` "
                    f"which is different from the `config_type` parameter in `{filename}` which is `{config_type}`!"
                )
        config_cls = get_module_config_class(config["name"], registry_type=config_type)
        if config_cls is None:
            config_cls = cls
        config = config_cls.from_dict(config, **kwargs)
        return config

    @classmethod
    def from_dict(cls, dict_config: Union[Dict, DictConfig], **kwargs):
        """
        Load config from a dict-like object. Nested configs are also properly converted to their classes if possible.
        """
        # Update config parameters with kwargs
        dict_config.update(**kwargs)

        for k, v in dict_config.items():
            if isinstance(v, Dict) and "name" in v and "config_type" in v:
                config_cls = get_module_config_class(v["name"], v["config_type"])
                if config_cls is not None:
                    dict_config[k] = config_cls.from_dict(v)

        # Remove class vars to avoid TypeError (unexpected argument)
        [dict_config.pop(k, None) for k in CONFIG_CLASS_VARS]

        config = cls(**{k: v for k, v in dict_config.items() if k in cls.__annotations__})  # noqa

        return config

    def save(
        self,
        save_dir: Union[str, os.PathLike],
        filename: str,
        subfolder: Optional[str] = None,
        skip_none_fields: Optional[bool] = True,
    ):
        """
        Save the `*config.yaml` file to a local path

        Args:
             save_dir: Save directory path
             filename: Config file name
             subfolder: Subfolder to save the config file
             skip_none_fields (bool): Whether to skip saving None values or not
        """
        subfolder = subfolder or ""
        config = self.dict()

        if skip_none_fields:
            # exclude None items
            config = {k: v for k, v in config.items() if v is not None}

        # make and save to directory
        os.makedirs(os.path.join(save_dir, subfolder), exist_ok=True)
        save_path = os.path.join(save_dir, subfolder, filename)
        OmegaConf.save(config, save_path)

        return save_path

    def push_to_hub(
        self,
        repo_id: str,
        filename: str,
        subfolder: Optional[str] = None,
        repo_type: Optional[str] = "model",
        skip_none_fields: Optional[bool] = True,
        private: Optional[bool] = False,
        commit_message: Optional[str] = None,
    ):
        """
        Push the config file to the hub

        Args:
            repo_id (str): Repo name or id on the Hub
            filename (str): config file name
            subfolder (str): subfolder to save the config
            repo_type (str): Type of the repo e.g, model, dataset, space
            skip_none_fields (bool): Whether to skip saving None values or not
            private (bool): Whether the repo type should be private or not (ignored if the repo exists)
            commit_message (str): Push commit message
        """
        path_in_repo = f"{subfolder}/{filename}" if subfolder else filename
        subfolder = subfolder or ""

        # create remote repo
        create_repo(repo_id, repo_type=repo_type, private=private, exist_ok=True)
        # save to tmp and prepare for push
        cache_path = tempfile.mkdtemp()
        config_path = self.save(cache_path, filename=filename, subfolder=subfolder, skip_none_fields=skip_none_fields)
        # push to hub
        if commit_message is None:
            commit_message = f"Hezar: Upload {filename}"
        upload_file(
            path_or_fileobj=config_path,
            path_in_repo=path_in_repo,
            repo_id=repo_id,
            repo_type=repo_type,
            commit_message=commit_message,
        )
        logger.log_upload_success(name=f"{self.__class__.__name__}()", target_path=os.path.join(repo_id, path_in_repo))


@dataclass
class ModelConfig(Config):
    """
    Base dataclass for all model configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.MODEL)


@dataclass
class PreprocessorConfig(Config):
    """
    Base dataclass for all preprocessor configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.PREPROCESSOR)


@dataclass
class DatasetConfig(Config):
    """
    Base dataclass for all dataset configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.DATASET)
    task: Union[TaskType, List[TaskType]] = field(
        default=None, metadata={"help": "Name of the task(s) this dataset is built for"}
    )


@dataclass
class EmbeddingConfig(Config):
    """
    Base dataclass for all embedding configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.EMBEDDING)


@dataclass
class MetricConfig(Config):
    """
    Base dataclass config for all metric configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.METRIC)
    output_keys: Union[List, Tuple] = None
    n_decimals: int = 4


@dataclass
class TrainerConfig(Config):
    """
    Base dataclass for all trainer configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.TRAINER)
    task: Union[str, TaskType] = None
    device: str = "cuda"
    init_weights_from: str = None
    dataset_config: Union[DatasetConfig, Dict] = None
    num_dataloader_workers: int = 0
    seed: int = 42
    optimizer: str = None
    learning_rate: float = 2e-5
    weight_decay: float = 0.0
    scheduler: str = None
    batch_size: int = None
    eval_batch_size: int = None
    use_amp: bool = False
    evaluate_with_generate: bool = True
    metrics: List[Union[str, MetricConfig]] = None
    num_epochs: int = None
    save_freq: int = 1
    checkpoints_dir: str = None
    logs_dir: str = "logs"

    def __post_init__(self):
        super().__post_init__()
        if self.task is None:
            raise ValueError("The parameter `task` is required for `TrainerConfig`!")
        if self.task not in list(TaskType):
            raise ValueError(
                f"Invalid task `{self.task}` passed to `TrainerConfig`. "
                f"Available options are {[t.value for t in list(TaskType)]}",
            )



import inspect
import os
import random
from typing import Any, Callable, Dict, Mapping, Tuple, Union

import numpy as np
import torch
from huggingface_hub import create_repo, hf_hub_download, upload_file
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from ..configs import TrainerConfig
from ..constants import (
    DEFAULT_DATASET_CONFIG_FILE,
    DEFAULT_TRAINER_CONFIG_FILE,
    DEFAULT_TRAINER_CSV_LOG_FILE,
    DEFAULT_TRAINER_SUBFOLDER,
    HEZAR_CACHE_DIR,
    TQDM_BAR_FORMAT,
    TaskType,
)
from ..data.datasets import Dataset
from ..models import Model
from ..preprocessors import Preprocessor, PreprocessorsContainer
from ..utils import Logger, sanitize_params_for_fn
from .metrics_handlers import (
    AudioClassificationMetricsHandler,
    Image2TextMetricHandler,
    MetricsHandler,
    SequenceLabelingMetricsHandler,
    SpeechRecognitionMetricsHandler,
    TextClassificationMetricsHandler,
    TextGenerationMetricsHandler,
)
from .trainer_utils import CSVLogger, write_to_tensorboard


logger = Logger(__name__)

optimizers = {
    "adam": torch.optim.Adam,
    "adamw": torch.optim.AdamW,
    "sgd": torch.optim.SGD,
}
lr_schedulers = {
    "reduce_on_plateau": torch.optim.lr_scheduler.ReduceLROnPlateau,
    "cosine_lr": torch.optim.lr_scheduler.CosineAnnealingLR,
}

task_to_metrics_handlers_mapping = {
    TaskType.TEXT_CLASSIFICATION: TextClassificationMetricsHandler,
    TaskType.SEQUENCE_LABELING: SequenceLabelingMetricsHandler,
    TaskType.IMAGE2TEXT: Image2TextMetricHandler,
    TaskType.SPEECH_RECOGNITION: SpeechRecognitionMetricsHandler,
    TaskType.AUDIO_CLASSIFICATION: AudioClassificationMetricsHandler,
    TaskType.TEXT_GENERATION: TextGenerationMetricsHandler,

}


class Trainer:
    """
    Base trainer class for training all Hezar models and all tasks. Usually you can use this class as-is, but for special
    cases you can also override any of the core methods in your own custom Trainer.

    Args:
        model ([`Model`] or `torch.nn.Module`): The main model to train and evaluate
        config (TrainerConfig): Training configuration and parameters
        train_dataset (Dataset): Train dataset
        eval_dataset (Dataset): Evaluation dataset
        data_collator: Collate function, usually included in the dataset object itself
        preprocessor: Preprocessor object(s)
        metrics_handler: Optional metrics handler
        optimizer (optim.Optimizer): Model optimizer
        lr_scheduler: Optional learning-rate scheduler

    """

    trainer_subfolder = DEFAULT_TRAINER_SUBFOLDER
    trainer_config_file = DEFAULT_TRAINER_CONFIG_FILE
    trainer_csv_log_file = DEFAULT_TRAINER_CSV_LOG_FILE
    dataset_config_file = DEFAULT_DATASET_CONFIG_FILE
    default_optimizer = "adam"
    default_lr_scheduler = None

    def __init__(
        self,
        model: Model = None,
        config: TrainerConfig = None,
        train_dataset: Dataset = None,
        eval_dataset: Dataset = None,
        data_collator: Callable = None,
        preprocessor: Union[Preprocessor, PreprocessorsContainer] = None,
        metrics_handler: MetricsHandler = None,
        optimizer: torch.optim.Optimizer = None,
        lr_scheduler=None,
    ):
        self.config = config

        self.device, self.device_type = self._prepare_device_and_type()
        self.autocast_dtype = torch.bfloat16 if self.device_type == "cpu" else torch.float16
        self.scaler = torch.cuda.amp.GradScaler(enabled=self.config.use_amp and self.device_type == "cuda")

        self._set_seed(self.config.seed)

        self.model = self._prepare_model(model)
        self.model.preprocessor = preprocessor

        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset
        self.data_collator = data_collator or self.train_dataset.data_collator
        self.train_dataloader, self.eval_dataloader = self._prepare_dataloaders()

        self.optimizer, self.lr_scheduler = self._prepare_optimizers(optimizer, lr_scheduler)

        self.metrics_handler = metrics_handler or self._prepare_metrics_handler()

        self.tensorboard = SummaryWriter(log_dir=self.config.logs_dir)
        self.csv_logger = CSVLogger(logs_dir=self.config.logs_dir, csv_filename=self.trainer_csv_log_file)

    def _prepare_device_and_type(self) -> Tuple[str, str]:
        device = self.config.device if "cuda" in self.config.device and torch.cuda.is_available() else "cpu"
        device_type = "cuda" if "cuda" in device else "cpu"
        return device, device_type

    @staticmethod
    def _set_seed(seed) -> None:
        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)

    def _prepare_model(self, model: Model) -> Model:
        """
        Download the model from HuggingFace Hub if `init_weights_from` is given in the config. Load the model to the
        device and return it.
        """
        if model is None:
            raise ValueError("`model` must be given to the Trainer!")
        hub_path = self.config.init_weights_from
        if hub_path is not None:
            if os.path.isdir(hub_path):
                model_path = os.path.join(hub_path, model.model_filename)
            else:
                model_path = hf_hub_download(
                    hub_path,
                    filename=model.model_filename,
                    cache_dir=HEZAR_CACHE_DIR,
                    resume_download=True,
                )
            model.load_state_dict(torch.load(model_path, map_location="cpu"))
        model.to(self.device)
        return model

    def _prepare_dataloaders(self) -> Tuple[DataLoader, Union[DataLoader, None]]:
        """
        Set up data loaders (train/eval) and return them.

        Returns:
             A tuple of train and eval dataloaders
        """
        if self.train_dataset is not None:
            train_dataloader = DataLoader(
                dataset=self.train_dataset,
                batch_size=self.config.batch_size,
                collate_fn=self.data_collator,
                num_workers=self.config.num_dataloader_workers,
                drop_last=True,
                shuffle=True,
            )
        else:
            raise ValueError("Cannot create train dataloader because `train_dataset` is not given!")
        if self.eval_dataset is not None:
            eval_dataloader = DataLoader(
                dataset=self.eval_dataset,
                batch_size=self.config.eval_batch_size or self.config.batch_size,
                collate_fn=self.data_collator,
                num_workers=self.config.num_dataloader_workers,
                drop_last=True,
                shuffle=True,
            )
        else:
            logger.warning(
                "Cannot create eval dataloader because `eval_dataset` is not given to the Trainer! "
                "Setting eval_dataloader to None..."
            )
            eval_dataloader = None

        return train_dataloader, eval_dataloader

    def _prepare_optimizers(self, optimizer: torch.optim.Optimizer = None, lr_scheduler=None):
        """
        Set up the optimizer and lr scheduler if they're not already given

        Args:
            optimizer: If None do nothing and return it, otherwise build it using the train config
            lr_scheduler: If None do nothing and return it, otherwise build it using the train config

        Returns:
            Optimizer and scheduler
        """
        if optimizer is None:
            optimizer_type = self.config.optimizer or self.default_optimizer
            optimizer = optimizers[optimizer_type](
                self.model.parameters(),
                lr=self.config.learning_rate,
                weight_decay=self.config.weight_decay,
            )

            if lr_scheduler is None:
                scheduler_name = self.config.scheduler or self.default_lr_scheduler
                if scheduler_name is None:
                    lr_scheduler = None
                else:
                    lr_scheduler = lr_schedulers[scheduler_name](optimizer, verbose=True)
        return optimizer, lr_scheduler

    def _prepare_metrics_handler(self):
        metrics_handler_cls = task_to_metrics_handlers_mapping[self.config.task]
        metrics_handler = metrics_handler_cls(
            metrics=self.config.metrics,
            trainer=self,
        )
        return metrics_handler

    def prepare_input_batch(self, input_batch) -> Dict[str, torch.Tensor]:
        """
        Every operation required to prepare the inputs for model forward like moving to device, permutations, etc.
        Args:
            input_batch: Raw input batch from the dataloader

        Returns:
            The proper input batch required by model forward
        """
        # cast to device
        input_batch = {k: v.to(self.device) if isinstance(v, torch.Tensor) else v for k, v in input_batch.items()}
        return input_batch

    def amp_context_manager(self):
        """
        A smart context manager for mixed precision.

        Returns:
            A torch autocast context manager
        """
        return torch.autocast(device_type=self.device_type, dtype=self.autocast_dtype, enabled=self.config.use_amp)

    def forward(self, input_batch):
        """
        Perform model forward on the input batch

        In special cases, one can override this method in their desired trainer.

        Args:
            input_batch: Input batch

        Returns:
            Model outputs
        """
        if isinstance(input_batch, torch.Tensor):
            outputs = self.model(input_batch)
        elif isinstance(input_batch, Dict):
            forward_inputs = sanitize_params_for_fn(self.model.forward, input_batch)
            outputs = self.model(**forward_inputs)
        else:
            raise ValueError(
                f"`input_batch` must be a tensor or a dict-like object containing key/value pairs of tensors, "
                f"but got {type(input_batch)}"
            )
        if not isinstance(outputs, Dict):
            raise ValueError(f"Model outputs must be dict-like not `{type(outputs)}`")

        return outputs

    def compute_loss(self, model_outputs: Dict, labels: torch.Tensor, **kwargs) -> torch.Tensor:
        """
        Compute loss from model outputs

        This method must be implemented in other trainers.

        Args:
            model_outputs: Logits from model outputs
            labels: Ground truth labels

        Returns:
            The loss tensor
        """
        compute_loss_inputs = sanitize_params_for_fn(self.model.compute_loss, model_outputs, **kwargs)
        compute_loss_inputs["labels"] = labels

        loss = self.model.compute_loss(**compute_loss_inputs)

        return loss

    def training_step(self, input_batch: Dict[str, torch.Tensor]) -> Dict[str, Any]:
        """
        Train one batch of data and return loss and model outputs

        Args:
            input_batch: A batch of inputs to train

        Returns:
            Train step outputs including loss, logits, etc.
        """
        with self.amp_context_manager():
            outputs = self.forward(input_batch)
            loss = self.compute_loss(outputs, **input_batch)

        self.scaler.scale(loss).backward()
        self.scaler.step(self.optimizer)
        self.scaler.update()
        self.optimizer.zero_grad()

        outputs["loss"] = loss.item() if isinstance(loss, torch.Tensor) else loss

        return outputs

    def evaluation_step(self, input_batch: Dict[str, torch.Tensor]) -> Dict[str, Any]:
        """
        Evaluate one batch of data and return loss and model outputs

        Args:
            input_batch: A batch of inputs to evaluate

        Returns:
            Evaluation step outputs including loss, logits, etc.
        """
        with self.amp_context_manager():
            outputs = self.forward(input_batch)
            loss = self.compute_loss(outputs, **input_batch)
            if self.model.is_generative and self.config.evaluate_with_generate:
                generate_inputs = sanitize_params_for_fn(self.model.generate, input_batch)
                generated_ids = self.model.generate(**generate_inputs)
                outputs["logits"] = generated_ids

        outputs["loss"] = loss.item() if isinstance(loss, torch.Tensor) else loss

        return outputs

    def inner_training_loop(self, epoch_num: int):
        """
        Train the model for one epoch on the whole train dataset and verbose live metric values in the progress bar

        Args:
            epoch_num: Number of the current epoch

        Returns:
            Metrics averages through the full iteration
        """
        losses_sum = 0.0
        self.model.train()
        with tqdm(
            self.train_dataloader,
            unit="batch",
            desc=f"Epoch: {epoch_num}/{self.config.num_epochs} ",
            bar_format=TQDM_BAR_FORMAT,
            ascii=" #",
        ) as iterator:
            for step, input_batch in enumerate(iterator):
                input_batch = self.prepare_input_batch(input_batch)
                # Training on one batch
                outputs = self.training_step(input_batch)
                losses_sum += outputs["loss"]
                # Gather outputs for metrics
                avg_loss = losses_sum / (step + 1)
                iterator.set_postfix(loss=avg_loss)

        return {"loss": avg_loss}

    def evaluate(self):
        """
        Evaluates the model on the whole eval dataset and verbose live metric values in the progress bar

        Returns:
            Evaluation results
        """
        self.metrics_handler.tracker.reset()
        self.model.eval()
        with tqdm(
            self.eval_dataloader,
            unit="batch",
            desc="Evaluating... ",
            bar_format=TQDM_BAR_FORMAT,
            ascii=" #",
        ) as iterator:
            with torch.inference_mode():
                for step, input_batch in enumerate(iterator):
                    input_batch = self.prepare_input_batch(input_batch)
                    # Evaluation on one batch
                    outputs = self.evaluation_step(input_batch)
                    logits = outputs["logits"].detach().cpu().numpy()
                    labels = input_batch["labels"].detach().cpu().numpy()
                    # Compute metrics
                    evaluation_results = self.metrics_handler.compute_metrics(logits, labels)
                    evaluation_results["loss"] = outputs["loss"]
                    # Gather outputs for metrics
                    self.metrics_handler.tracker.update(evaluation_results)
                    iterator.set_postfix(**self.metrics_handler.tracker.avg())

        return self.metrics_handler.tracker.avg()

    def train(self):
        """
        The full training process like training, evaluation, logging and saving model checkpoints.
        """
        for epoch in range(1, self.config.num_epochs + 1):
            print()
            training_results = self.inner_training_loop(epoch)
            evaluation_results = self.evaluate()
            if self.lr_scheduler is not None:
                self.lr_scheduler.step(evaluation_results["loss"])

            current_step_logs = {
                "step": epoch,
                "training_results": training_results,
                "evaluation_results": evaluation_results,
            }

            self.log(current_step_logs)

            # maybe save checkpoint
            if epoch % self.config.save_freq == 0:
                ckpt_save_path = os.path.join(self.config.checkpoints_dir, str(epoch))
                self.save(ckpt_save_path)

    def log(self, logs: Dict[str, Any]):
        """
        Log metrics results
        """
        step = logs["step"]
        training_results = logs["training_results"]
        evaluation_results = logs["evaluation_results"]

        train_logs = {f"train.{metric_name}": value for metric_name, value in training_results.items()}
        evaluation_logs = {f"evaluation.{metric_name}": value for metric_name, value in evaluation_results.items()}

        # Log to tensorboard
        write_to_tensorboard(self.tensorboard, train_logs, step)
        write_to_tensorboard(self.tensorboard, evaluation_logs, step)

        # Log to CSV
        self.csv_logger.write({**train_logs, **evaluation_logs}, step)

    def save(
        self,
        path: str,
        config_filename=None,
        model_filename=None,
        model_config_filename=None,
        subfolder=None,
        dataset_config_file=None,
    ):
        """
        Save the trainer and relevant files to a path.

        Files to save are train config, model weights, model config, preprocessor files and preprocessor config.

        Args:
            path: A directory to save everything
            config_filename: Config filename
            model_filename: Model file name
            model_config_filename: Model config file name
            subfolder: Optional sub-folder
            dataset_config_file: Dataset config filename
        """
        config_filename = config_filename or self.trainer_config_file
        subfolder = subfolder or self.trainer_subfolder
        dataset_config_file = dataset_config_file or self.dataset_config_file

        self.config.save(path, filename=config_filename, subfolder=subfolder)
        self.model.save(path, filename=model_filename, config_filename=model_config_filename)
        self.train_dataset.config.save(path, filename=dataset_config_file, subfolder=subfolder)

    def push_to_hub(
        self,
        repo_id: str,
        config_filename: str = None,
        push_model: bool = True,
        push_logs: bool = True,
        model_filename: str = None,
        model_config_filename: str = None,
        subfolder: str = None,
        dataset_config_filename: str = None,
        commit_message: str = None,
        private: bool = False,
    ):
        """
        Push everything to the Hub

        Args:
            repo_id: Path to hub
            config_filename: Trainer config file name
            push_model: Whether to push the model or not
            push_logs: Whether to push training logs or not
            model_filename: Model file name
            model_config_filename: Model config file name
            subfolder: Path to Trainer files
            dataset_config_filename: Dataset config file name
            commit_message: Commit message for the push
            private: Whether to create a private repo if it doesn't exist already
        """
        config_filename = config_filename or self.trainer_config_file
        subfolder = subfolder or self.trainer_subfolder
        dataset_config_file = dataset_config_filename or self.dataset_config_file

        # create remote repo
        create_repo(repo_id, repo_type="model", exist_ok=True, private=private)

        if not commit_message:
            commit_message = "Hezar: Upload training files"

        # upload train files
        self.config.push_to_hub(
            repo_id,
            filename=config_filename,
            subfolder=subfolder,
            private=private,
            commit_message=commit_message,
        )
        self.train_dataset.config.push_to_hub(
            repo_id,
            filename=dataset_config_file,
            subfolder=subfolder,
            private=private,
            commit_message=commit_message
        )

        # upload model
        if push_model:
            self.model.push_to_hub(
                repo_id,
                filename=model_filename,
                config_filename=model_config_filename,
                commit_message=commit_message,
                private=private,
            )

        if push_logs:
            upload_file(
                path_or_fileobj=self.csv_logger.save_path,
                path_in_repo=os.path.join(self.trainer_subfolder, self.trainer_csv_log_file),
                repo_id=repo_id,
                commit_message=commit_message,
            )



from ..registry import register_preprocessor  # noqa
from .preprocessor import Preprocessor, PreprocessorConfig, PreprocessorsContainer  # noqa
from .audio_feature_extractor import AudioFeatureExtractor, AudioFeatureExtractorConfig
from .image_processor import ImageProcessor, ImageProcessorConfig
from .text_normalizer import TextNormalizer, TextNormalizerConfig
from .tokenizers import *



from .tokenizer import Tokenizer, TokenizerConfig
from .bpe import BPEConfig, BPETokenizer
from .sentencepiece_bpe import SentencePieceBPEConfig, SentencePieceBPETokenizer
from .sentencepiece_unigram import (
    SentencePieceUnigramConfig,
    SentencePieceUnigramTokenizer,
)
from .wordpiece import WordPieceConfig, WordPieceTokenizer



from dataclasses import dataclass, field

from ....configs import ModelConfig


@dataclass
class DecoderConfig(ModelConfig):
    name = "beit_roberta_decoder"
    is_decoder: bool = True
    add_cross_attention: bool = True
    attention_probs_dropout_prob: float = 0.1
    bos_token_id: int = 0
    eos_token_id: int = 2
    classifier_dropout: float = None
    gradient_checkpointing: bool = False
    hidden_act: str = "gelu"
    hidden_dropout_prob: float = 0.1
    hidden_size: int = 768
    initializer_range: int = 0.02
    intermediate_size: int = 3072
    layer_norm_eps: float = 1e-12
    max_position_embeddings: int = 514
    num_attention_heads: int = 12
    num_hidden_layers: int = 12
    pad_token_id: int = 1
    position_embedding_type: str = "absolute"
    type_vocab_size: int = 1
    use_cache: bool = True
    vocab_size: int = 42000


@dataclass
class EncoderConfig(ModelConfig):
    name = "beit_roberta_encoder"
    vocab_size = 8192
    hidden_size = 768
    num_hidden_layers = 12
    num_attention_heads = 12
    intermediate_size = 3072
    hidden_act = "gelu"
    hidden_dropout_prob = 0.0
    attention_probs_dropout_prob = 0.0
    initializer_range = 0.02
    layer_norm_eps = 1e-12
    image_size = 224
    patch_size = 16
    num_channels = 3
    use_mask_token = False
    use_absolute_position_embeddings = False
    use_relative_position_bias = False
    use_shared_relative_position_bias = False
    layer_scale_init_value = 0.1
    drop_path_rate = 0.1
    use_mean_pooling = True
    out_indices = [3, 5, 7, 11]
    pool_scales = [1, 2, 3, 6]
    use_auxiliary_head = True
    auxiliary_loss_weight = 0.4
    auxiliary_channels = 256
    auxiliary_num_convs = 1
    auxiliary_concat_input = False
    semantic_loss_ignore_index = 255


@dataclass
class GenerationConfig:
    bos_token_id: int = 0
    decoder_start_token_id: int = 0
    early_stopping: bool = True
    eos_token_id: int = 2
    length_penalty: float = 2.0
    max_length: int = 64
    no_repeat_ngram_size: int = 3
    num_beams: int = 4
    pad_token_id: int = 1


@dataclass
class BeitRobertaImage2TextConfig(ModelConfig):
    name = "beit_roberta_image2text"
    encoder: EncoderConfig = field(default_factory=EncoderConfig)
    decoder: DecoderConfig = field(default_factory=DecoderConfig)
    generation: GenerationConfig = field(default_factory=GenerationConfig)



from dataclasses import dataclass

from ..configs import MetricConfig
from ..constants import Backends, MetricType
from ..registry import register_metric
from ..utils import is_backend_available
from .metric import Metric


if is_backend_available(Backends.ROUGE):
    from rouge_score import rouge_scorer, scoring

_DESCRIPTION = "Rouge estimation. Commonly used for Text Summarization"

_required_backends = [
    Backends.ROUGE,
]


@dataclass
class ROUGEConfig(MetricConfig):
    """
    Configuration class for ROUGE metric.

    Args:
        name (MetricType): The type of metric, ROUGE in this case.
        use_stemmer (bool): Flag to enable stemming when computing ROUGE.
        use_aggregator (bool): Flag to enable score aggregation for multiple references.
        multi_ref (bool): Flag to indicate if multiple references are present.
        output_keys (tuple): Keys to filter the metric results for output.
    """
    name = MetricType.ROUGE
    use_stemmer: bool = False
    use_aggregator: bool = True
    multi_ref: bool = True
    output_keys: tuple = ("rouge1", "rouge2", "rougeL", "rougeLsum",)


@register_metric("rouge", config_class=ROUGEConfig, description=_DESCRIPTION)
class ROUGE(Metric):
    """
    ROUGE metric for evaluating text summarization using `rouge_score`.

    Args:
        config (ROUGEConfig): Metric configuration object.
        **kwargs: Extra configuration parameters passed as kwargs to update the `config`.
    """
    required_backends = _required_backends

    def __init__(self, config: ROUGEConfig, **kwargs):
        super().__init__(config=config, **kwargs)
        rouge_types = ["rouge1", "rouge2", "rougeL", "rougeLsum"]
        self.scorer = rouge_scorer.RougeScorer(
            rouge_types=rouge_types,
            use_stemmer=self.config.use_stemmer,
        )

    def compute(
        self,
        predictions=None,
        targets=None,
        use_aggregator=None,
        n_decimals=None,
        output_keys=None,
        **kwargs,
    ):
        """
        Computes the ROUGE scores for the given predictions against targets.

        Args:
            predictions: Predicted summaries.
            targets: Ground truth summaries.
            use_aggregator (bool): Flag to enable score aggregation for multiple references.
            n_decimals (int): Number of decimals for the final score.
            output_keys (tuple): Filter the output keys.

        Returns:
            dict: A dictionary of the metric results, with keys specified by `output_keys`.
        """
        aggregator = scoring.BootstrapAggregator()

        for ref, pred in zip(targets, predictions):
            if self.config.multi_ref:
                score = self.scorer.score_multi(ref, pred)
            else:
                score = self.scorer.score(ref, pred)

            aggregator.add_scores(score)

        results = aggregator.aggregate()
        for key in results:
            results[key] = results[key].mid.fmeasure

        if output_keys:
            results = {k: round(v, 4) for k, v in results.items() if k in output_keys}

        return results



from typing import List, Union

import numpy as np
import torch

from ....constants import Backends
from ....registry import register_model
from ....utils import is_backend_available
from ...model import Model
from ...model_outputs import Image2TextOutput
from .vit_roberta_image2text_config import ViTRobertaImage2TextConfig


if is_backend_available(Backends.TRANSFORMERS):
    from transformers import (
        GenerationConfig,
        RobertaConfig,
        RobertaForCausalLM,
        VisionEncoderDecoderModel,
        ViTConfig,
        ViTModel,
    )

if is_backend_available(Backends.PILLOW):
    from PIL import Image

_required_backends = [Backends.TRANSFORMERS, Backends.TOKENIZERS, Backends.PILLOW]


@register_model("vit_roberta_image2text", config_class=ViTRobertaImage2TextConfig)
class ViTRobertaImage2Text(Model):
    """
    ViT + RoBERTa for image to text
    """

    is_generative = True
    required_backends = _required_backends
    image_processor = "image_processor"
    tokenizer_name = "bpe_tokenizer"
    loss_fn_name = "cross_entropy"

    def __init__(self, config: ViTRobertaImage2TextConfig, **kwargs):
        super().__init__(config, **kwargs)
        encoder = ViTModel(config=ViTConfig(**self.config.encoder))
        decoder = RobertaForCausalLM(config=RobertaConfig(**self.config.decoder))
        self.vit_roberta = VisionEncoderDecoderModel(encoder=encoder, decoder=decoder)

    def forward(
        self,
        pixel_values,
        decoder_input_ids=None,
        decoder_attention_mask=None,
        encoder_outputs=None,
        past_key_values=None,
        decoder_inputs_embeds=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        **kwargs,
    ):
        outputs = self.vit_roberta(
            pixel_values=pixel_values,
            decoder_input_ids=decoder_input_ids,
            decoder_attention_mask=decoder_attention_mask,
            encoder_outputs=encoder_outputs,
            past_key_values=past_key_values,
            decoder_inputs_embeds=decoder_inputs_embeds,
            labels=None,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )

        return outputs

    def compute_loss(self, logits: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        loss = self.criterion(logits.reshape(-1, self.vit_roberta.decoder.config.vocab_size), labels.reshape(-1))
        return loss

    def generate(self, pixel_values, generation_config=None, **kwargs):
        tokenizer = self.preprocessor["bpe_tokenizer"]
        if generation_config is None:
            generation_config = self.config.dict()["generation"]
            generation_config["decoder_start_token_id"] = tokenizer.pad_token_id
        generation_config = GenerationConfig(**generation_config)
        outputs = self.vit_roberta.generate(inputs=pixel_values, generation_config=generation_config, **kwargs)

        return outputs

    def preprocess(self, inputs: Union[List[str], List[np.ndarray], List["Image"], List[torch.Tensor]], **kwargs):
        image_processor = self.preprocessor[self.image_processor]
        processed_outputs = image_processor(inputs, **kwargs)
        return processed_outputs

    def post_process(self, model_outputs, **kwargs):
        tokenizer = self.preprocessor[self.tokenizer_name]
        decoded_outputs = tokenizer.decode(model_outputs.cpu().numpy().tolist())
        outputs = [Image2TextOutput(text=text) for text in decoded_outputs]
        return outputs



from dataclasses import dataclass, field

import torch
from datasets import load_dataset

from ...configs import DatasetConfig
from ...constants import TaskType
from ...preprocessors import Tokenizer
from ...registry import register_dataset
from ...utils import Logger
from ..data_collators import TextPaddingDataCollator
from .dataset import Dataset


logger = Logger(__name__)


@dataclass
class TextClassificationDatasetConfig(DatasetConfig):
    """
    Configuration class for text classification datasets.

    Args:
        path (str): Path to the dataset.
        tokenizer_path (str): Path to the tokenizer file.
        label_field (str): Field name for labels in the dataset.
        text_field (str): Field name for text in the dataset.
        max_length (int): Maximum length of text.
    """

    name = "text_classification"
    task: TaskType = TaskType.TEXT_CLASSIFICATION
    path: str = None
    tokenizer_path: str = None
    label_field: str = None
    text_field: str = None
    max_length: int = None


@register_dataset("text_classification", config_class=TextClassificationDatasetConfig)
class TextClassificationDataset(Dataset):
    """
    A text classification dataset class.
    As of now this class is intended for datasets existing on the Hub!

    Args:
        config (TextClassificationDatasetConfig): Dataset config object.
        split: Which split to use.
        **kwargs: Extra config parameters to assign to the original config.
    """

    def __init__(self, config: TextClassificationDatasetConfig, split=None, **kwargs):
        """
        Initializes a new TextClassificationDataset instance.

        Args:
            config (TextClassificationDatasetConfig): The configuration object for the dataset.
            split: Dataset split, defaults to None.
            **kwargs: Additional keyword arguments.

        """
        super().__init__(config, **kwargs)
        self.dataset = self._load(split)
        self._extract_labels()
        self.tokenizer = self._build_tokenizer()
        self.data_collator = TextPaddingDataCollator(
            tokenizer=self.tokenizer,
            max_length=self.config.max_length,
        )

    def _load(self, split):
        """
        Load the dataset.

        Args:
            split: Dataset split.

        Returns:
            The whole dataset.

        """
        # TODO: In case we want to make this class work on other types like csv, json, etc. we have to do it here.
        dataset = load_dataset(self.config.path, split=split, cache_dir=self.cache_dir)
        return dataset

    def _build_tokenizer(self):
        """
        Build the tokenizer.

        Returns:
            Tokenizer: The tokenizer.

        """
        if self.config.tokenizer_path:
            tokenizer = Tokenizer.load(self.config.tokenizer_path)
        else:
            logger.warning(
                "This dataset requires a tokenizer to work. Provide it in config as `tokenizer_path` "
                "or set it manually as `dataset.tokenizer = your_tokenizer` after building the dataset."
            )
            tokenizer = None
        return tokenizer

    def _extract_labels(self):
        """
        Extract label names, ids and build dictionaries.
        """
        labels_list = self.dataset.features[self.config.label_field].names
        self.id2label = self.config.id2label = {k: str(v) for k, v in dict(enumerate(labels_list)).items()}
        self.label2id = self.config.label2id = {v: k for k, v in self.id2label.items()}
        self.num_labels = self.config.num_labels = len(labels_list)

    def __len__(self):
        """
        Returns the length of the dataset.

        Returns:
            int: The length of the dataset.

        """
        return len(self.dataset)

    def __getitem__(self, index):
        """
        Tokenize inputs and return a dict containing ids, masks, labels, etc.

        Args:
            index: Sample index.

        Returns:
            dict: The input data.

        """
        text = self.dataset[index][self.config.text_field]
        label = self.dataset[index][self.config.label_field]
        inputs = self.tokenizer(
            text,
            return_tensors="pt",
            truncation_strategy="longest_first",
            padding="longest",
            return_attention_mask=True,
        )
        label_idx = torch.tensor([label], dtype=torch.long)  # noqa
        inputs["labels"] = label_idx

        return inputs



from dataclasses import dataclass
from typing import Iterable, Union

from ..configs import MetricConfig
from ..constants import Backends, MetricType
from ..registry import register_metric
from ..utils import is_backend_available
from .metric import Metric


if is_backend_available(Backends.NLTK):
    from nltk.translate.bleu_score import corpus_bleu

_required_backends = [
    Backends.NLTK,
]


@dataclass
class BLEUConfig(MetricConfig):
    """
    Configuration class for BLEU metric.

    args:
        name (MetricType): The type of metric, BLEU in this case.
        output_keys (tuple): Keys to filter the metric results for output.
    """
    name = MetricType.BLEU
    output_keys: tuple = ("bleu",)


@register_metric("bleu", config_class=BLEUConfig)
class BLEU(Metric):
    """
    BLEU metric for evaluating text generation models like translation, summarization, etc.
    """
    required_backends = _required_backends

    def __init__(self, config: BLEUConfig, **kwargs):
        super().__init__(config=config, **kwargs)

    def compute(
        self,
        predictions: Union[Iterable[str], str] = None,
        targets: Union[Iterable[str], str] = None,
        weights=(0.25, 0.25, 0.25, 0.25),
        n_decimals=None,
        output_keys=None,
        **kwargs,
    ):
        """
        Computes the BLEU score for the given predictions against targets.

        Args:
            predictions (Union[Iterable[str], str]): Predicted sentences or tokens.
            targets (Union[Iterable[str], str]): Ground truth sentences or tokens.
            weights (tuple): Weights for n-gram precision, default is (0.25, 0.25, 0.25, 0.25).
            n_decimals (int): Number of decimals for the final score.
            output_keys (tuple): Filter the output keys.

        Returns:
            dict: A dictionary of the metric results, with keys specified by `output_keys`.
        """
        n_decimals = n_decimals or self.config.n_decimals
        output_keys = output_keys or self.config.output_keys

        predictions = [x.split() if isinstance(x, str) else x for x in predictions]
        targets = [x.split() if isinstance(x, str) else x for x in targets]

        score = corpus_bleu(targets, predictions, weights=weights)

        results = {"bleu": round(float(score), n_decimals)}

        if output_keys:
            results = {k: v for k, v in results.items() if k in output_keys}

        return results



"""
A BERT model for text classification built using HuggingFace Transformers
"""
from typing import Dict, List, Union

import torch
from torch import nn

from ....constants import Backends
from ....registry import register_model
from ....utils import is_backend_available
from ...model import Model
from ...model_outputs import TextClassificationOutput
from .bert_text_classification_config import BertTextClassificationConfig


if is_backend_available(Backends.TRANSFORMERS):
    from transformers import BertConfig, BertModel

_required_backends = [
    Backends.TRANSFORMERS,
    Backends.TOKENIZERS,
]


@register_model(model_name="bert_text_classification", config_class=BertTextClassificationConfig)
class BertTextClassification(Model):
    """
    A standard 🤗Transformers Bert model for text classification

    Args:
        config: The whole model config including arguments needed for the inner 🤗Transformers model.
    """

    required_backends = _required_backends
    tokenizer_name = "wordpiece_tokenizer"
    skip_keys_on_load = [
        "model.embeddings.position_ids",  # For older versions
        "bert.embeddings.position_ids",
        "model.bert.embeddings.position_ids",
    ]

    def __init__(self, config: BertTextClassificationConfig, **kwargs):
        super().__init__(config, **kwargs)
        self.bert = BertModel(self._build_inner_config())
        classifier_dropout = (
            self.config.classifier_dropout
            if self.config.classifier_dropout is not None
            else self.config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)
        self.classifier = nn.Linear(self.config.hidden_size, self.config.num_labels)

    def _build_inner_config(self):
        if self.config.num_labels is None and self.config.id2label is None:
            raise ValueError("Both `num_labels` and `id2label` are None. Please provide at least one of them!")
        if self.config.id2label and self.config.num_labels is None:
            self.config.num_labels = len(self.config.id2label)
        bert_config = BertConfig(**self.config)
        return bert_config

    def compute_loss(self, logits: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        criterion = nn.CrossEntropyLoss()
        loss = criterion(logits.view(-1, self.config.num_labels), labels.view(-1))
        return loss

    def forward(
        self,
        token_ids,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        **kwargs,
    ) -> Dict:
        lm_outputs = self.bert(
            token_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=True,
        )
        pooled_output = lm_outputs[1]
        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)
        outputs = {
            "logits": logits,
            "hidden_states": lm_outputs.hidden_states,
            "attentions": lm_outputs.attentions,
        }
        return outputs

    def preprocess(self, inputs: Union[str, List[str]], **kwargs):
        if isinstance(inputs, str):
            inputs = [inputs]
        if "text_normalizer" in self.preprocessor:
            normalizer = self.preprocessor["text_normalizer"]
            inputs = normalizer(inputs)
        tokenizer = self.preprocessor[self.tokenizer_name]
        inputs = tokenizer(inputs, return_tensors="pt", device=self.device)
        return inputs

    def post_process(self, model_outputs: dict, top_k=1):
        output_logits = model_outputs["logits"]
        outputs = []
        for logits in output_logits:
            probs = logits.softmax(dim=-1)
            scores, label_ids = probs.sort(descending=True)
            row = []
            for i, (score, label_id) in enumerate(zip(scores, label_ids)):
                if i == top_k:
                    break
                label_str = self.config.id2label[label_id.item()]
                score = score.item()
                row.append(TextClassificationOutput(label=label_str, score=score))
            outputs.append(row)
        return outputs



"""
Define all model outputs here
"""
from dataclasses import asdict, dataclass
from typing import Dict, List, Optional


@dataclass
class ModelOutput:
    """
    Base class for all models' prediction outputs (`model.predict()`/`model.post_process()` outputs).

    Note that prediction outputs must all be a list of `ModelOutput` objects since we consider only batch inferences.

    The helper functions in the class enable it to be treated as a mapping or a dict object.
    """

    def dict(self):
        return asdict(self)

    def __str__(self):
        return str({k: v for k, v in self.dict().items() if v is not None})

    def __repr__(self):
        return str(self)

    def __getitem__(self, item):
        try:
            return self.dict()[item]
        except KeyError:
            raise AttributeError(f"`{self.__class__.__name__}` has no attribute `{item}`!")

    def __len__(self):
        return len(self.dict())

    def __iter__(self):
        return iter(self.dict())

    def keys(self):
        return list(self.dict().keys())

    def values(self):
        return list(self.dict().values())

    def items(self):
        return self.dict().items()


@dataclass(repr=False)
class LanguageModelingOutput(ModelOutput):
    token: Optional[int] = None
    sequence: Optional[str] = None
    token_id: Optional[str] = None
    score: Optional[float] = None


@dataclass(repr=False)
class TextClassificationOutput(ModelOutput):
    label: Optional[str] = None
    score: Optional[float] = None


@dataclass(repr=False)
class SequenceLabelingOutput(ModelOutput):
    token: Optional[List[List[str]]] = None
    label: Optional[List[List[str]]] = None
    start: Optional[int] = None
    end: Optional[int] = None
    score: Optional[List[List[float]]] = None


@dataclass(repr=False)
class TextGenerationOutput(ModelOutput):
    text: Optional[str] = None


@dataclass(repr=False)
class SpeechRecognitionOutput(ModelOutput):
    text: Optional[str] = None
    chunks: Optional[List[Dict]] = None


@dataclass(repr=False)
class Image2TextOutput(ModelOutput):
    text: Optional[str] = None



from dataclasses import dataclass, field
from typing import List

from ...constants import DEFAULT_TOKENIZER_CONFIG_FILE, DEFAULT_TOKENIZER_FILE, Backends
from ...registry import register_preprocessor
from ...utils import is_backend_available
from .tokenizer import Tokenizer, TokenizerConfig


if is_backend_available(Backends.TOKENIZERS):
    from tokenizers import Tokenizer as HFTokenizer
    from tokenizers import decoders, models, trainers

_required_backends = [
    Backends.TOKENIZERS,
]


@dataclass
class WordPieceConfig(TokenizerConfig):
    name = "wordpiece_tokenizer"
    max_length: int = 512
    truncation_strategy: str = "longest_first"
    truncation_direction: str = "right"
    stride: int = 0
    padding_strategy: str = "longest"
    padding_direction: str = "right"
    pad_to_multiple_of: int = 0
    pad_token: str = "[PAD]"
    unk_token: str = "[UNK]"
    sep_token: str = "[SEP]"
    cls_token: str = "[CLS]"
    mask_token: str = "[MASK]"
    pad_token_type_id: int = 0
    additional_special_tokens: List[str] = None
    wordpieces_prefix: str = "##"
    vocab_size: int = 30000
    min_frequency: int = 2
    limit_alphabet: int = 1000
    initial_alphabet: list = field(default_factory=list)
    show_progress: bool = True


@register_preprocessor("wordpiece_tokenizer", config_class=WordPieceConfig)
class WordPieceTokenizer(Tokenizer):
    """
    A standard WordPiece tokenizer using 🤗HuggingFace Tokenizers

    Args:
        config: Preprocessor config for the tokenizer
        **kwargs: Extra/manual config parameters
    """

    required_backends = _required_backends

    tokenizer_filename = DEFAULT_TOKENIZER_FILE
    tokenizer_config_filename = DEFAULT_TOKENIZER_CONFIG_FILE
    token_ids_name = "token_ids"

    def __init__(self, config, tokenizer_file=None, **kwargs):
        super().__init__(config, tokenizer_file=tokenizer_file, **kwargs)

    def build(self):
        tokenizer = HFTokenizer(models.WordPiece(unk_token=self.config.unk_token))  # noqa
        tokenizer.decoder = decoders.WordPiece(self.config.wordpieces_prefix)  # noqa
        return tokenizer

    def train(self, files: List[str], **train_kwargs):
        """Train the model using the given files"""
        self.config.update(train_kwargs)

        trainer = trainers.WordPieceTrainer(
            vocab_size=self.config.vocab_size,
            min_frequency=self.config.min_frequency,
            limit_alphabet=self.config.limit_alphabet,
            initial_alphabet=self.config.initial_alphabet,
            special_tokens=self.config.special_tokens,
            show_progress=self.config.show_progress,
            continuing_subword_prefix=self.config.wordpieces_prefix,
        )
        if isinstance(files, str):
            files = [files]
        self._tokenizer.train(files, trainer=trainer)

    def train_from_iterator(self, dataset: List[str], **train_kwargs):
        """Train the model using the given files"""
        self.config.update(train_kwargs)

        trainer = trainers.WordPieceTrainer(
            vocab_size=self.config.vocab_size,
            min_frequency=self.config.min_frequency,
            limit_alphabet=self.config.limit_alphabet,
            initial_alphabet=self.config.initial_alphabet,
            special_tokens=self.config.special_tokens,
            show_progress=self.config.show_progress,
            continuing_subword_prefix=self.config.wordpieces_prefix,
        )
        self._tokenizer.train_from_iterator(dataset, trainer=trainer, length=len(dataset))



from .beit_roberta import BeitRobertaImage2Text, BeitRobertaImage2TextConfig
from .crnn import CRNNImage2Text, CRNNImage2TextConfig
from .trocr import TrOCRImage2Text, TrOCRImage2TextConfig
from .vit_gpt2 import ViTGPT2Image2Text, ViTGPT2Image2TextConfig
from .vit_roberta import ViTRobertaImage2Text, ViTRobertaImage2TextConfig



import gzip
import json
import os
import shutil
from typing import Union

import omegaconf

from .logging import Logger


logger = Logger(__name__)

__all__ = [
    "gunzip",
    "load_yaml_config",
    "load_json_config",
]


def load_yaml_config(path: Union[str, os.PathLike]):
    """
    Load yaml file using omegaconf
    """
    config = omegaconf.OmegaConf.load(path)
    return config


def load_json_config(path: Union[str, os.PathLike]):
    """
    Load json config file
    """
    with open(path) as f:
        config = json.load(f)
    return config


def gunzip(src_path, dest_path):
    """
    Unzip a .gz file from `src_path` and extract to `dest_path`
    Args:
        src_path: Path to .gz file
        dest_path: Path to the destination file

    Returns:

    """
    with gzip.open(src_path, "rb") as f_in:
        with open(dest_path, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
    logger.debug(f"Extracted {src_path} to {dest_path}")



from typing import List, Union

import numpy as np
import torch

from ....constants import Backends
from ....registry import register_model
from ....utils import is_backend_available, load_audio_files
from ...model import Model
from ...model_outputs import SpeechRecognitionOutput
from .whisper_speech_recognition_config import WhisperSpeechRecognitionConfig


if is_backend_available(Backends.TRANSFORMERS):
    from transformers import WhisperConfig, WhisperForConditionalGeneration


_required_backends = [
    Backends.TRANSFORMERS,
    Backends.TOKENIZERS,
    Backends.LIBROSA,
]


@register_model("whisper_speech_recognition", config_class=WhisperSpeechRecognitionConfig)
class WhisperSpeechRecognition(Model):
    """
    Whisper model for automatic speech recognition
    """

    is_generative = True
    required_backends = _required_backends
    feature_extractor_name = "whisper_feature_extractor"
    tokenizer_name = "whisper_bpe_tokenizer"
    loss_fn_name = "cross_entropy"

    def __init__(self, config: WhisperSpeechRecognitionConfig, **kwargs):
        super().__init__(config, **kwargs)
        self.whisper = WhisperForConditionalGeneration(WhisperConfig(**self.config))

    def forward(
        self,
        input_features,
        attention_mask=None,
        decoder_input_ids=None,
        decoder_attention_mask=None,
        head_mask=None,
        decoder_head_mask=None,
        cross_attn_head_mask=None,
        encoder_outputs=None,
        past_key_values=None,
        decoder_inputs_embeds=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        **kwargs,
    ):
        outputs = self.whisper(
            input_features=input_features,
            attention_mask=attention_mask,
            decoder_input_ids=decoder_input_ids,
            decoder_attention_mask=decoder_attention_mask,
            head_mask=head_mask,
            decoder_head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            encoder_outputs=encoder_outputs,
            past_key_values=past_key_values,
            decoder_inputs_embeds=decoder_inputs_embeds,
            labels=None,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )

        return outputs

    def compute_loss(self, logits: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        loss = self.criterion(logits.view(-1, self.config.vocab_size), labels.view(-1))
        return loss

    def generate(
        self,
        input_features,
        forced_decoder_ids=None,
        generation_config=None,
        logits_processor=None,
        stopping_criteria=None,
        prefix_allowed_tokens_fn=None,
        synced_gpus=None,
        return_timestamps=None,
        task=None,
        language=None,
        is_multilingual=None,
        prompt_ids=None,
        **kwargs,
    ):
        generation_outputs = self.whisper.generate(
            inputs=input_features,
            generation_config=generation_config,
            logits_processor=logits_processor,
            stopping_criteria=stopping_criteria,
            prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
            synced_gpus=synced_gpus,
            return_timestamps=return_timestamps,
            task=task,
            language=language,
            is_multilingual=is_multilingual,
            prompt_ids=prompt_ids,
            forced_decoder_ids=forced_decoder_ids,
            **kwargs,
        )
        return generation_outputs

    def prepare_inputs_for_generation(
        self,
        decoder_input_ids,
        past_key_values=None,
        use_cache=None,
        encoder_outputs=None,
        attention_mask=None,
        **kwargs,
    ):
        return self.whisper.prepare_inputs_for_generation(
            decoder_input_ids=decoder_input_ids,
            past_key_values=past_key_values,
            use_cache=use_cache,
            encoder_outputs=encoder_outputs,
            attention_mask=attention_mask,
            **kwargs,
        )

    def get_encoder(self):
        return self.whisper.get_encoder()

    def get_decoder(self):
        return self.whisper.get_decoder()

    def resize_token_embeddings(self, new_num_tokens: int) -> torch.nn.Embedding:
        new_embeddings = self.whisper.resize_token_embeddings(new_num_tokens)
        return new_embeddings

    def get_output_embeddings(self):
        return self.whisper.proj_out

    def set_output_embeddings(self, new_embeddings):
        self.whisper.set_output_embeddings(new_embeddings)

    def get_input_embeddings(self) -> torch.nn.Module:
        return self.whisper.get_input_embeddings()

    def freeze_encoder(self):
        self.whisper.freeze_encoder()

    def preprocess(self, inputs: Union[str, np.ndarray, List[np.ndarray], List[str]], **kwargs):
        if isinstance(inputs, str) or (isinstance(inputs, List) and isinstance(inputs[0], str)):
            inputs = load_audio_files(inputs)

        tokenizer = self.preprocessor[self.tokenizer_name]
        feature_extractor = self.preprocessor[self.feature_extractor_name]

        forced_decoder_ids = tokenizer.get_decoder_prompt_ids(language="persian", task="transcribe")
        inputs = feature_extractor(inputs, sampling_rate=self.config.sampling_rate, return_tensors="pt")
        inputs["forced_decoder_ids"] = forced_decoder_ids
        return inputs

    def post_process(self, model_outputs, **kwargs):
        tokenizer = self.preprocessor[self.tokenizer_name]
        transcripts = tokenizer.decode(model_outputs, decode_with_timestamps=True, skip_special_tokens=True)
        outputs = [SpeechRecognitionOutput(text=transcript) for transcript in transcripts]
        return outputs



from .bert import BertSequenceLabeling, BertSequenceLabelingConfig
from .distilbert import DistilBertSequenceLabeling, DistilBertSequenceLabelingConfig
from .roberta import RobertaSequenceLabeling, RobertaSequenceLabelingConfig



from typing import Dict, List, Union

import torch

from ....constants import Backends
from ....registry import register_model
from ....utils import is_backend_available
from ...model import Model
from ...model_outputs import TextGenerationOutput
from .t5_text_generation_config import T5TextGenerationConfig


if is_backend_available(Backends.TRANSFORMERS):
    from transformers import T5Config, T5ForConditionalGeneration

_required_backends = [
    Backends.TRANSFORMERS,
    Backends.TOKENIZERS,
]


@register_model("t5_text_generation", config_class=T5TextGenerationConfig)
class T5TextGeneration(Model):
    """
    T5 for text to text generation
    """

    is_generative = True
    required_backends = _required_backends
    tokenizer_name = "sentencepiece_unigram_tokenizer"
    loss_fn_name = "cross_entropy"

    def __init__(self, config: T5TextGenerationConfig, **kwargs):
        super().__init__(config=config, **kwargs)

        self.t5 = T5ForConditionalGeneration(T5Config(**self.config))

    def forward(
        self,
        token_ids,
        labels=None,
        attention_mask=None,
        decoder_input_ids=None,
        decoder_attention_mask=None,
        head_mask=None,
        decoder_head_mask=None,
        cross_attn_head_mask=None,
        encoder_outputs=None,
        past_key_values=None,
        inputs_embeds=None,
        decoder_inputs_embeds=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        **kwargs,
    ) -> Dict:

        if labels is not None and decoder_input_ids is None and decoder_inputs_embeds is None:
            # get decoder inputs from shifting lm labels to the right
            decoder_input_ids = self._shift_right(labels)

        outputs = self.t5(
            input_ids=token_ids,
            attention_mask=attention_mask,
            decoder_input_ids=decoder_input_ids,
            decoder_attention_mask=decoder_attention_mask,
            head_mask=head_mask,
            decoder_head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            encoder_outputs=encoder_outputs,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            decoder_inputs_embeds=decoder_inputs_embeds,
            labels=None,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )

        return dict(outputs)

    def _shift_right(self, input_ids):
        return self.t5._shift_right(input_ids)

    def compute_loss(self, logits: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        labels = labels.clone()
        labels[labels == self.config.pad_token_id] = -100
        loss = self.criterion(logits.view(-1, logits.size(-1)), labels.view(-1))
        return loss

    def generate(self, token_ids, attention_mask=None, **kwargs):
        # TODO Merge kwargs into generation config so users can control generation from kwargs
        model_inputs = {"input_ids": token_ids, "attention_mask": attention_mask}
        generation_kwargs = {"min_length": self.config.min_length, "max_length": self.config.max_length}
        output_ids = self.t5.generate(**model_inputs, **generation_kwargs)
        return output_ids

    def preprocess(self, inputs: Union[str, List[str]], prefix=None):
        if isinstance(inputs, str):
            inputs = [inputs]
        prefix = prefix or self.config.input_prefix
        if prefix:
            inputs = [f"{prefix}{x}" for x in inputs]
        tokenizer = self.preprocessor[self.tokenizer_name]
        inputs = tokenizer(inputs, return_tensors="pt", device=self.device)
        return inputs

    def post_process(self, generated_ids: torch.Tensor, **kwargs):
        tokenizer = self.preprocessor[self.tokenizer_name]
        decoded_outputs = tokenizer.decode(generated_ids.cpu().numpy().tolist())
        outputs = [TextGenerationOutput(text=text) for text in decoded_outputs]
        return outputs



"""
A RoBERTa Language Model (HuggingFace Transformers) wrapped by a Hezar Model class
"""
from typing import List, Union

import torch

from ....constants import Backends
from ....models import Model
from ....registry import register_model
from ....utils import is_backend_available
from ...model_outputs import LanguageModelingOutput
from .roberta_lm_config import RobertaLMConfig


if is_backend_available(Backends.TRANSFORMERS):
    from transformers import RobertaConfig, RobertaForMaskedLM

_required_backends = [
    Backends.TRANSFORMERS,
    Backends.TOKENIZERS,
]


@register_model("roberta_lm", config_class=RobertaLMConfig)
class RobertaLM(Model):
    required_backends = _required_backends
    tokenizer_name = "bpe_tokenizer"
    skip_keys_on_load = ["model.embeddings.position_ids", "roberta.embeddings.position_ids"]  # For older versions
    loss_fn_name = "cross_entropy"

    def __init__(self, config, **kwargs):
        super().__init__(config=config, **kwargs)
        self.roberta_mlm = RobertaForMaskedLM(RobertaConfig(**self.config))

    def forward(
        self,
        token_ids,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        output_attentions=None,
        output_hidden_states=None,
        **kwargs,
    ):
        outputs = self.roberta_mlm(
            input_ids=token_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            labels=None,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )
        outputs["token_ids"] = token_ids

        return outputs

    def compute_loss(self, logits: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        loss = self.criterion(logits.view(-1, self.config.vocab_size), labels.view(-1))
        return loss

    def preprocess(self, inputs: Union[str, List[str]], **kwargs):
        if isinstance(inputs, str):
            inputs = [inputs]

        tokenizer = self.preprocessor[self.tokenizer_name]

        for text in inputs:
            if tokenizer.mask_token not in text:
                raise ValueError(f"The input must have a `{tokenizer.mask_token}` token!")

        inputs = tokenizer(inputs, return_tensors="pt", device=self.device)
        return inputs

    def post_process(self, model_outputs: dict, top_k=1):
        output_logits = model_outputs.get("logits")
        token_ids = model_outputs.get("token_ids")

        tokenizer = self.preprocessor[self.tokenizer_name]
        mask_token_id = tokenizer.mask_token_id

        unfilled_token_ids = token_ids.cpu().numpy().copy()
        outputs = []
        for batch_i, logits in enumerate(output_logits):
            masked_index = torch.nonzero(token_ids[batch_i] == mask_token_id, as_tuple=False).flatten()  # noqa
            if len(masked_index) > 1:
                raise ValueError(
                    f"Can't handle multiple `{tokenizer.mask_token}` tokens in the input for {self.__class__.__name__}!"
                )
            logits = logits[masked_index]
            probs = logits.softmax(dim=-1)
            probs, top_fill_token_ids = probs.topk(top_k)
            if top_k != 1:
                probs = probs.squeeze()
                top_fill_token_ids = top_fill_token_ids.squeeze()

            row = []
            for i, (prob, token_id) in enumerate(zip(probs, top_fill_token_ids)):
                candidate = unfilled_token_ids[batch_i].copy()
                candidate[masked_index.item()] = token_id
                row.append(
                    LanguageModelingOutput(
                        token=tokenizer.decode([token_id.item()])[0].strip(),
                        sequence=tokenizer.decode(candidate.tolist())[0],
                        token_id=token_id.item(),
                        score=prob.item(),
                    )
                )
            outputs.append(row)
        return outputs



"""
Direct importing from hezar's root is no longer supported nor recommended since version 0.33.0. The following is just a
workaround for backward compatibility. Any class, functions, etc. must be imported from its main submodule under hezar.
"""
import warnings


__version__ = "0.33.1"


def _warn_on_import(name: str, submodule: str):
    warnings.warn(
        f"Importing {name} from hezar root is deprecated and will be removed soon. "
        f"Please use `from {submodule} import {name}`"
    )


def __getattr__(name: str):
    if name == "Model":
        from hezar.models import Model
        _warn_on_import(name, "hezar.models")
        return Model
    elif name == "Dataset":
        from .data import Dataset
        _warn_on_import(name, "hezar.data")
        return Dataset
    elif name == "Trainer":
        from .trainer import Trainer
        _warn_on_import(name, "hezar.trainer")
        return Trainer
    elif name == "Embedding":
        from .embeddings import Embedding
        _warn_on_import(name, "hezar.embeddings")
        return Embedding
    elif name == "Preprocessor":
        from .preprocessors import Preprocessor
        _warn_on_import(name, "hezar.preprocessors")
        return Preprocessor
    elif name == "Metric":
        from .metrics import Metric
        _warn_on_import(name, "hezar.metrics")
        return Metric
    elif "Config" in name:
        from .configs import Config
        _warn_on_import(name, "hezar.configs")
        return Config


__all__ = [
    "Config",
    "Model",
    "Dataset",
    "Trainer",
    "Preprocessor",
    "Embedding",
    "Metric",
]



from dataclasses import dataclass, field

from ....configs import ModelConfig
from ....constants import TaskType


@dataclass
class BertSequenceLabelingConfig(ModelConfig):
    name = "bert_sequence_labeling"
    task: str = TaskType.SEQUENCE_LABELING
    num_labels: int = None
    id2label: dict = None
    vocab_size: int = 42000
    hidden_size: int = 768
    num_hidden_layers: int = 12
    num_attention_heads: int = 12
    intermediate_size: int = 3072
    hidden_act: str = "gelu"
    hidden_dropout_prob: float = 0.1
    attention_probs_dropout_prob: float = 0.1
    max_position_embeddings: int = 512
    type_vocab_size: int = 2
    initializer_range: float = 0.02
    layer_norm_eps: float = 1e-12
    pad_token_id: int = 0
    position_embedding_type: str = "absolute"
    use_cache: bool = True
    classifier_dropout: float = None
    prediction_skip_tokens: list = field(default_factory=lambda: ["[SEP]", "[CLS]"])



from ...registry import register_dataset  # noqa
from .dataset import Dataset, DatasetConfig  # noqa
from .ocr_dataset import OCRDataset, OCRDatasetConfig
from .sequence_labeling_dataset import SequenceLabelingDataset, SequenceLabelingDatasetConfig
from .text_classification_dataset import TextClassificationDataset, TextClassificationDatasetConfig
from .text_summarization_dataset import TextSummarizationDataset, TextSummarizationDatasetConfig



import torch
from torch import nn

from ....registry import register_model
from ...model import Model
from ...model_outputs import Image2TextOutput
from .crnn_decode_utils import ctc_decode
from .crnn_image2text_config import CRNNImage2TextConfig


@register_model("crnn_image2text", config_class=CRNNImage2TextConfig)
class CRNNImage2Text(Model):
    """
    A robust CRNN model for character level OCR based on the original paper.
    """

    is_generative = True
    image_processor = "image_processor"
    loss_fn_name = "ctc"

    def __init__(self, config: CRNNImage2TextConfig, **kwargs):
        super().__init__(config=config, **kwargs)
        self.cnn = nn.Sequential(
            ConvBlock(self.config.n_channels, 64, 3, 1, 1),
            nn.MaxPool2d(kernel_size=2, stride=2),
            ConvBlock(64, 128, 3, 1, 1),
            nn.MaxPool2d(kernel_size=2, stride=2),
            ConvBlock(128, 256, 3, 1, 1),
            ConvBlock(256, 256, 3, 1, 1),
            nn.MaxPool2d(kernel_size=(2, 1)),
            ConvBlock(256, 512, 3, 1, 1, batch_norm=True),
            ConvBlock(512, 512, 3, 1, 1, batch_norm=True),
            nn.MaxPool2d(kernel_size=(2, 1)),
            ConvBlock(512, 512, 3, 1, 1),
        )
        # map CNN to sequence
        self.map2seq = nn.Linear(self.config.map2seq_in_dim, self.config.map2seq_out_dim)
        # RNN
        self.rnn1 = nn.LSTM(self.config.map2seq_out_dim, self.config.rnn_dim, bidirectional=True)
        self.rnn2 = nn.LSTM(2 * self.config.rnn_dim, self.config.rnn_dim, bidirectional=True)
        # classifier
        self.classifier = nn.Linear(2 * self.config.rnn_dim, len(self.config.id2label))

    def forward(self, pixel_values, **kwargs):
        # CNN block
        x = self.cnn(pixel_values)
        # reformat array
        batch, channel, height, width = x.size()
        x = x.view(batch, channel * height, width)
        x = x.permute(2, 0, 1)
        x = self.map2seq(x)
        x, _ = self.rnn1(x)
        x, _ = self.rnn2(x)
        x = self.classifier(x)
        x = nn.functional.log_softmax(x, 2)
        outputs = {"logits": x}
        return outputs

    def compute_loss(self, logits: torch.Tensor, labels: torch.Tensor):
        batch_size = logits.size(1)
        labels_lengths = torch.LongTensor([len(t) for t in labels])
        labels = labels.flatten()
        input_lengths = torch.LongTensor([logits.size(0)] * batch_size)

        loss = self.criterion(logits, labels, input_lengths, labels_lengths) / batch_size

        return loss

    def generate(self, pixel_values, **kwargs):
        logits = self(pixel_values)["logits"]
        output_ids = ctc_decode(logits, blank=self.config.blank_id)
        return output_ids

    def preprocess(self, inputs, **kwargs):
        image_processor = self.preprocessor[self.image_processor]
        processed_outputs = image_processor(inputs, **kwargs)
        return processed_outputs

    def post_process(self, model_outputs: torch.Tensor, **kwargs):
        outputs = []
        generated_ids = model_outputs.cpu().numpy().tolist()
        for decoded_ids in generated_ids:
            chars = [self.config.id2label[id_] for id_ in decoded_ids]
            text = "".join(chars)
            if self.config.reverse_prediction_text:
                text = text[::-1]
            outputs.append(Image2TextOutput(text=text))
        return outputs


class ConvBlock(nn.Module):
    def __init__(self, input_channel, output_channel, kernel_sizes, strides, paddings, batch_norm: bool = False):
        super(ConvBlock, self).__init__()
        self.do_batch_norm = batch_norm
        self.conv = nn.Conv2d(input_channel, output_channel, kernel_sizes, strides, paddings)
        self.bn = nn.BatchNorm2d(output_channel)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        if self.do_batch_norm:
            x = self.bn(x)
        x = self.relu(x)
        return x



import os
from typing import List, Optional, Union

from torch.utils.data import Dataset as TorchDataset

from ...builders import build_dataset
from ...configs import DatasetConfig
from ...constants import (
    DEFAULT_DATASET_CONFIG_FILE,
    HEZAR_CACHE_DIR,
    Backends,
    RepoType,
    SplitType,
)
from ...utils import verify_dependencies


class Dataset(TorchDataset):
    """
    Base class for all datasets in Hezar.

    Args:
        config (DatasetConfig): The configuration object for the dataset.
        **kwargs: Additional keyword arguments.

    Attributes:
        required_backends (List[Union[str, Backends]]): List of required backends for the dataset.
        config_filename (str): Default dataset config file name.
        cache_dir (str): Default cache directory for the dataset.

    """
    required_backends: List[Union[str, Backends]] = None
    config_filename = DEFAULT_DATASET_CONFIG_FILE
    cache_dir = HEZAR_CACHE_DIR

    def __init__(self, config: DatasetConfig, **kwargs):
        """
        Initializes a new dataset instance.

        Args:
            config (DatasetConfig): The configuration object for the dataset.
            **kwargs: Additional keyword arguments.

        """
        verify_dependencies(self, self.required_backends)
        self.config = config.update(kwargs)
        self.preprocessor = None
        self.data_collator = None

    def __len__(self):
        """
        Returns the length of the dataset.

        Raises:
            NotImplementedError: This method must be implemented in derived classes.

        """
        raise NotImplementedError

    def __getitem__(self, index):
        """
        Gets a specific item from the dataset.

        Args:
            index: Index of the item to retrieve.

        Raises:
            NotImplementedError: This method must be implemented in derived classes.

        """
        raise NotImplementedError

    @classmethod
    def load(
        cls,
        hub_path: Union[str, os.PathLike],
        config_filename: Optional[str] = None,
        split: Optional[Union[str, SplitType]] = None,
        **kwargs,
    ) -> "Dataset":
        """
        Load the dataset from a hub path.

        Args:
            hub_path (Union[str, os.PathLike]): Path to dataset from hub or locally.
            config_filename (Optional[str]): Dataset config file name.
            split (Optional[Union[str, SplitType]]): Dataset split, defaults to "train".
            **kwargs: Config parameters as keyword arguments.

        Returns:
            Dataset: An instance of the loaded dataset.

        """
        split = split or "train"
        config_filename = config_filename or cls.config_filename
        dataset_config = DatasetConfig.load(hub_path, filename=config_filename, repo_type=RepoType.DATASET)
        dataset_config.path = hub_path
        dataset = build_dataset(dataset_config.name, config=dataset_config, split=split, **kwargs)
        return dataset
